#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Mapping, List

import random
import numpy as np

from .state import FockState

from piquasso.instructions import gates

from piquasso._math.decompositions import euler

from piquasso.api.instruction import Instruction
from piquasso.api.result import Result

from piquasso._math.indices import get_index_in_fock_space
from piquasso._math.fock import (
    get_single_mode_displacement_operator,
    get_creation_operator,
    get_annihilation_operator,
    operator_basis,
    get_single_mode_squeezing_operator,
    get_single_mode_cubic_phase_operator,
    get_fock_space_basis,
)

from ..calculations import (
    calculate_state_index_matrix_list,
    calculate_interferometer_helper_indices,
    calculate_index_list_for_appling_interferometer,
    get_projection_operator_indices,
)


def vacuum(state: FockState, instruction: Instruction, shots: int) -> Result:
    state.reset()

    return Result(state=state)


def passive_linear(
    state: FockState, instruction: gates._PassiveLinearGate, shots: int
) -> Result:
    operator: np.ndarray = instruction._get_passive_block(
        state._connector, state._config
    )

    _apply_passive_linear(state, operator, instruction.modes)

    return Result(state=state)


def _apply_passive_linear(state, interferometer, modes):
    subspace_transformations = _get_interferometer_on_fock_space(
        interferometer, state._config.cutoff, state._connector
    )

    _apply_passive_gate_matrix_to_state(state, subspace_transformations, modes)


def _apply_passive_gate_matrix_to_state(
    state: FockState,
    subspace_transformations: List[np.ndarray],
    modes: Tuple[int, ...],
) -> None:
    index_list = calculate_index_list_for_appling_interferometer(
        modes,
        state.d,
        state._config.cutoff,
    )

    state._density_matrix = _calculate_density_matrix_after_interferometer(
        state._density_matrix,
        subspace_transformations,
        index_list,
    )


def _calculate_density_matrix_after_interferometer(
    density_matrix: np.ndarray,
    subspace_transformations: List[np.ndarray],
    index_list: List[np.ndarray],
) -> np.ndarray:
    new_density_matrix = np.empty_like(density_matrix)

    for ket, indices_ket in enumerate(index_list):
        for bra, indices_bra in enumerate(index_list):
            for col_ket in range(indices_ket.shape[1]):
                for col_bra in range(indices_bra.shape[1]):
                    index = np.ix_(
                        indices_ket[:, col_ket],
                        indices_bra[:, col_bra].T,
                    )
                    new_density_matrix[index] = np.einsum(
                        "ij,jk,kl->il",
                        subspace_transformations[ket],
                        density_matrix[index],
                        subspace_transformations[bra].T.conj(),
                    )

    return new_density_matrix


def _get_interferometer_on_fock_space(interferometer, cutoff, connector):
    index_tuple = calculate_interferometer_helper_indices(
        d=len(interferometer),
        cutoff=cutoff,
    )

    return connector.calculate_interferometer_on_fock_space(interferometer, index_tuple)


def particle_number_measurement(
    state: FockState, instruction: Instruction, shots: int
) -> Result:
    reduced_state = state.reduced(instruction.modes)

    probability_map = reduced_state.fock_probabilities_map

    samples = random.choices(
        population=list(probability_map.keys()),
        weights=list(probability_map.values()),
        k=shots,
    )

    # NOTE: We choose the last sample for multiple shots.
    sample = samples[-1]

    normalization = _get_normalization(probability_map, sample)

    _project_to_subspace(
        state=state,
        subspace_basis=sample,
        modes=instruction.modes,
        normalization=normalization,
    )

    return Result(state=state, samples=samples)


def _get_normalization(
    probability_map: Mapping[Tuple[int, ...], float], sample: Tuple[int, ...]
) -> float:
    return 1 / probability_map[sample]


def _project_to_subspace(
    state: FockState,
    *,
    subspace_basis: Tuple[int, ...],
    modes: Tuple[int, ...],
    normalization: float,
) -> None:
    projected_density_matrix = _get_projected_density_matrix(
        state=state,
        subspace_basis=subspace_basis,
        modes=modes,
    )

    state._density_matrix = projected_density_matrix * normalization


def _get_projected_density_matrix(
    state: FockState, *, subspace_basis: Tuple[int, ...], modes: Tuple[int, ...]
) -> np.ndarray:
    new_density_matrix = state._get_empty()

    index = get_projection_operator_indices(
        state.d, state._config.cutoff, modes, subspace_basis
    )

    operator_index = np.ix_(index, index)

    new_density_matrix[operator_index] = state._density_matrix[operator_index]

    return new_density_matrix


def create(state: FockState, instruction: Instruction, shots: int) -> Result:
    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    operator = get_creation_operator(
        instruction.modes, space=space, config=state._config
    )

    state._density_matrix = operator @ state._density_matrix @ operator.transpose()

    return Result(state=state)


def annihilate(state: FockState, instruction: Instruction, shots: int) -> Result:
    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    operator = get_annihilation_operator(
        instruction.modes, space=space, config=state._config
    )

    state._density_matrix = operator @ state._density_matrix @ operator.transpose()

    return Result(state=state)


def kerr(state: FockState, instruction: Instruction, shots: int) -> Result:
    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    xi = instruction._all_params["xi"]

    mode = instruction.modes[0]

    for index, (basis, dual_basis) in operator_basis(space):
        number = basis[mode]
        dual_number = dual_basis[mode]

        coefficient = np.exp(1j * xi * (number**2 - dual_number**2))

        state._density_matrix[index] *= coefficient

    return Result(state=state)


def cubic_phase(state: FockState, instruction: Instruction, shots: int) -> Result:
    gamma = instruction._all_params["gamma"]

    matrix = get_single_mode_cubic_phase_operator(
        gamma=gamma,
        cutoff=state._config.cutoff,
        hbar=state._config.hbar,
        connector=state._connector,
    )
    _apply_active_gate_matrix_to_state(state, matrix, instruction.modes[0])

    return Result(state=state)


def _apply_active_gate_matrix_to_state(
    state: FockState,
    matrix: np.ndarray,
    mode: int,
) -> None:
    density_matrix = state._density_matrix

    state_index_matrix_list = calculate_state_index_matrix_list(
        state.d, state._config.cutoff, mode
    )
    density_matrix = _calculate_density_matrix_after_apply_active_gate(
        density_matrix, matrix, state_index_matrix_list
    )

    state._density_matrix = density_matrix


def _calculate_density_matrix_after_apply_active_gate(
    density_matrix, matrix, state_index_matrix_list
):
    """
    Applies the active gate matrix to the density matrix specified by
    `state_index_matrix_list`.
    """

    new_density_matrix = np.empty_like(density_matrix, dtype=density_matrix.dtype)

    cutoff = len(state_index_matrix_list)

    for ket in range(cutoff):
        state_index_matrix_ket = state_index_matrix_list[ket]
        limit_ket = state_index_matrix_ket.shape[0]
        sliced_matrix_ket = matrix[:limit_ket, :limit_ket]

        for bra in range(max(ket + 1, cutoff)):
            state_index_matrix_bra = state_index_matrix_list[bra]
            limit_bra = state_index_matrix_bra.shape[0]
            sliced_matrix_bra = matrix[:limit_bra, :limit_bra]

            for col_ket in range(state_index_matrix_ket.shape[1]):
                for col_bra in range(state_index_matrix_bra.shape[1]):
                    index = np.ix_(
                        state_index_matrix_ket[:, col_ket],
                        state_index_matrix_bra[:, col_bra].T,
                    )
                    partial_result = np.einsum(
                        "ij,jk,kl->il",
                        sliced_matrix_ket,
                        density_matrix[index],
                        sliced_matrix_bra.T.conj(),
                    )

                    new_density_matrix[index] = partial_result

                    if bra < ket:
                        # NOTE: We use the fact that the density matrix is self-adjoint.
                        index = np.ix_(
                            state_index_matrix_bra[:, col_bra],
                            state_index_matrix_ket[:, col_ket].T,
                        )

                        new_density_matrix[index] = partial_result.conj().T

    return new_density_matrix


def cross_kerr(state: FockState, instruction: Instruction, shots: int) -> Result:
    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    modes = instruction.modes
    xi = instruction._all_params["xi"]

    for index, (basis, dual_basis) in operator_basis(space):
        coefficient = np.exp(
            1j
            * xi
            * (
                basis[modes[0]] * basis[modes[1]]
                - dual_basis[modes[0]] * dual_basis[modes[1]]
            )
        )

        state._density_matrix[index] *= coefficient

    return Result(state=state)


def displacement(state: FockState, instruction: Instruction, shots: int) -> Result:
    r = instruction._all_params["r"]
    phi = instruction._all_params["phi"]
    mode = instruction.modes[0]

    matrix = get_single_mode_displacement_operator(
        r=r,
        phi=phi,
        connector=state._connector,
        cutoff=state._config.cutoff,
        complex_dtype=state._config.complex_dtype,
    )

    _apply_active_gate_matrix_to_state(state, matrix, mode=mode)

    return Result(state=state)


def squeezing(state: FockState, instruction: Instruction, shots: int) -> Result:
    r = instruction._all_params["r"]
    phi = instruction._all_params["phi"]
    mode = instruction.modes[0]

    matrix = get_single_mode_squeezing_operator(
        r=r,
        phi=phi,
        connector=state._connector,
        cutoff=state._config.cutoff,
        complex_dtype=state._config.complex_dtype,
    )

    _apply_active_gate_matrix_to_state(state, matrix, mode=mode)

    return Result(state=state)


def linear(
    state: FockState, instruction: gates._ActiveLinearGate, shots: int
) -> Result:
    connector = state._connector
    modes = instruction.modes

    np = connector.np

    passive_block = instruction._get_passive_block(state._connector, state._config)
    active_block = instruction._get_active_block(state._connector, state._config)

    symplectic = connector.block(
        [
            [passive_block, active_block],
            [np.conj(active_block), np.conj(passive_block)],
        ],
    )

    unitary_last, squeezings, unitary_first = euler(symplectic, connector)

    _apply_passive_linear(state, unitary_first, modes)

    for mode, r in zip(instruction.modes, squeezings):
        matrix = get_single_mode_squeezing_operator(
            r=r,
            phi=0.0,
            connector=state._connector,
            cutoff=state._config.cutoff,
            complex_dtype=state._config.complex_dtype,
        )
        _apply_active_gate_matrix_to_state(state, matrix, mode)

    _apply_passive_linear(state, unitary_last, modes)

    return Result(state=state)


def density_matrix_instruction(
    state: FockState, instruction: Instruction, shots: int
) -> Result:
    _add_occupation_number_basis(state, **instruction.params)

    return Result(state=state)


def _add_occupation_number_basis(
    state: FockState,
    *,
    ket: Tuple[int, ...],
    bra: Tuple[int, ...],
    coefficient: complex,
) -> None:
    index = get_index_in_fock_space(ket)
    dual_index = get_index_in_fock_space(bra)

    state._density_matrix[index, dual_index] = coefficient



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq
import pytest


@pytest.mark.monkey
def test_Interferometer_numpy_array_as_parameter(generate_unitary_matrix, tf):
    r = tf.Variable(0.01)
    d = 5
    interferometer = generate_unitary_matrix(d)

    simulator = pq.PureFockSimulator(
        d=d, config=pq.Config(cutoff=3), connector=pq.TensorflowConnector()
    )

    with pq.Program() as program:
        pq.Q(all) | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=r)

        pq.Q(all) | pq.Interferometer(interferometer)

    simulator.execute(program)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np
import piquasso as pq

for_all_connectors = pytest.mark.parametrize(
    "connector", [pq.NumpyConnector(), pq.JaxConnector()]
)


@for_all_connectors
def test_density_matrix_StateVector_ordering(connector):
    d = 3

    state_vectors = [
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
        [1, 1, 0],
        [1, 0, 1],
        [0, 1, 1],
        [1, 1, 1],
    ]

    simulator = pq.fermionic.PureFockSimulator(d=d, connector=connector)

    for i in range(len(state_vectors)):
        with pq.Program() as preparation:
            pq.Q(0, 1, 2) | pq.StateVector(state_vectors[i])

        state = simulator.execute(preparation).state

        density_matrix = state.density_matrix

        assert np.isclose(density_matrix[i, i], 1.0)


@for_all_connectors
def test_StateVector_raises_InvalidParameter_for_invalid_occupation_numbers(connector):
    d = 3

    simulator = pq.fermionic.PureFockSimulator(d=d, connector=connector)

    with pq.Program() as preparation:
        pq.Q(0, 1, 2) | pq.StateVector([1, 2])

    with pytest.raises(pq.api.exceptions.InvalidParameter) as error:
        simulator.execute(preparation)

    assert error.value.args[0] == (
        "Invalid initial state specified: "
        "instruction="
        "StateVector(occupation_numbers=(1, 2), coefficient=1.0, modes=(0, 1, 2))"
    )


@for_all_connectors
def test_state_vector_with_fock_amplitude_map_preparation(connector):
    amplitude_map = {(0,): 0.6, (1,): 0.8}

    with pq.Program() as program:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map)

    simulator = pq.fermionic.PureFockSimulator(d=1, connector=connector)

    state = simulator.execute(program).state

    assert np.allclose(state.state_vector, np.array([0.6, 0.8]))


@for_all_connectors
def test_state_vector_with_fock_amplitude_map_and_coefficient(connector):
    amplitude_map = {(0,): 0.6, (1,): 0.8}
    coefficient = 1 / np.sqrt(2)

    with pq.Program() as program_with_mul:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map) * coefficient

    with pq.Program() as program_with_param:
        pq.Q() | pq.StateVector(
            fock_amplitude_map=amplitude_map, coefficient=coefficient
        )

    simulator = pq.fermionic.PureFockSimulator(d=1, connector=connector)

    state_with_mul = simulator.execute(program_with_mul).state
    state_with_param = simulator.execute(program_with_param).state

    expected = coefficient * np.array([0.6, 0.8])

    assert np.allclose(state_with_mul.state_vector, expected)
    assert np.allclose(state_with_param.state_vector, expected)


@for_all_connectors
def test_StateVector_raises_InvalidParameter_for_invalid_fock_amplitude_map(connector):
    amplitude_map = {(0, 1): 0.6}

    simulator = pq.fermionic.PureFockSimulator(d=1, connector=connector)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map)

    with pytest.raises(pq.api.exceptions.InvalidParameter) as error:
        simulator.execute(program)

    assert error.value.args[0] == (
        "Invalid initial state specified: "
        "instruction="
        "StateVector(fock_amplitude_map={(0, 1): 0.6}, coefficient=1.0, modes=(0,))"
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numba as nb

import numpy as np

from .labudde import labudde
from .hessenberg import (
    transform_matrix_to_hessenberg,
    transform_matrix_to_hessenberg_loop,
)


@nb.njit(cache=True)
def _powtrace_from_charpoly(coeffs_labudde, pow):
    dim = coeffs_labudde.shape[0]

    coeffs = coeffs_labudde.flatten()

    if pow == 0:
        return np.array([dim], dtype=coeffs.dtype)

    traces = np.empty(pow, dtype=coeffs.dtype)

    element_offset = (dim - 1) * dim
    traces[0] = -coeffs[element_offset]

    kdx_max = pow if pow < dim else dim
    for idx in range(2, kdx_max + 1):
        element_offset2 = element_offset + idx - 1
        traces[idx - 1] = -coeffs[element_offset2] * idx

        for j in range(idx - 1, 0, -1):
            traces[idx - 1] -= coeffs[element_offset2 - j] * traces[j - 1]

    element_offset_coeffs = (dim - 1) * dim - 1
    if pow > dim:
        for idx in range(1, pow - dim + 1):
            element_offset = dim + idx - 1
            traces[element_offset] = 0.0

            for jdx in range(1, dim + 1):
                traces[element_offset] -= (
                    traces[element_offset - jdx] * coeffs[element_offset_coeffs + jdx]
                )

    return traces


@nb.njit(cache=True)
def calc_power_traces(A, pow_max):
    """
    Calculate the traces of $A^pow_max$, where `pow_max` is an integer and `A` is a
    square matrix.

    The trace is calculated from the coefficients of its characteristic polynomial.
    In the case that the power `pow_max` is above the size of the matrix we can use an
    optimization described in Appendix B of
    [arxiv:1805.12498](https://arxiv.org/pdf/1104.3769v1.pdf).
    """

    transform_matrix_to_hessenberg(A)

    coeffs_labudde = labudde(A)

    return _powtrace_from_charpoly(coeffs_labudde, pow_max)


@nb.njit(cache=True)
def calculate_power_traces_loop(
    cx_diag_elements,
    diag_elements,
    AZ,
    pow_max,
):
    r"""
    Call to calculate the power traces $Tr(mtx^j)~\forall~1\leq j\leq l$ for a squared
    complex matrix $mtx$ of dimensions $n\times n$ according to Eq. (3.26) of
    [arxiv:1805.12498](https://arxiv.org/pdf/1104.3769v1.pdf).
    """
    transform_matrix_to_hessenberg_loop(AZ, diag_elements, cx_diag_elements)

    coeffs_labudde = labudde(AZ)

    return _powtrace_from_charpoly(coeffs_labudde, pow_max)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_5050_beamsplitter():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1))

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 3)

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=3))
    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [0, 0.5, 0.5, 0, 0, 0],
    )


def test_beamsplitter():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1))

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.3454915, 0.6545085, 0.0, 0.0, 0.0],
    )


def test_beamsplitter_multiple_particles():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2)) * np.sqrt(1 / 8)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.08637288, 0.16362712, 0.32397979, 0.17929466, 0.24672554],
    )


def test_beamsplitter_leaves_vacuum_unchanged():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 0), bra=(0, 0)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 2
        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.25, 0.17274575, 0.32725425, 0.02984109, 0.11306356, 0.10709534],
    )


def test_multiple_beamsplitters():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1))

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 5)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 6, phi=1.5 * np.pi)

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.25, 0.75, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    )


def test_multiple_beamsplitters_with_multiple_particles():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 1, 1)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 1, 1)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 0, 2)) * np.sqrt(1 / 8)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 5)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 6, phi=1.5 * np.pi)

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.0625, 0.1875, 0.0, 0.0625, 0.1875, 0.109375, 0.15625, 0.234375],
    )


def test_phaseshift():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 2
        pq.Q() | pq.DensityMatrix(ket=(1, 1), bra=(1, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(1, 1)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(1, 1), bra=(0, 2)) * np.sqrt(1 / 8)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0) | pq.Phaseshifter(phi=np.pi / 3)

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.25, 0.0, 0.25, 0.5],
    )


def test_fourier():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 2
        pq.Q() | pq.DensityMatrix(ket=(1, 1), bra=(1, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(1, 1)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(1, 1), bra=(0, 2)) * np.sqrt(1 / 8)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0) | pq.Fourier()

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.25, 0.0, 0.25, 0.5],
    )


def test_mach_zehnder():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 2
        pq.Q() | pq.DensityMatrix(ket=(1, 1), bra=(1, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(1, 1)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(1, 1), bra=(0, 2)) * np.sqrt(1 / 8)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0, 1) | pq.MachZehnder(int_=np.pi / 3, ext=np.pi / 4)

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.1875, 0.0625, 0.60463966, 0.09690689, 0.04845345],
    )


def test_beamsplitters_and_phaseshifters_with_multiple_particles():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 1, 1)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 1, 1)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 0, 2)) * np.sqrt(1 / 8)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0) | pq.Phaseshifter(phi=np.pi / 3)
        pq.Q(1) | pq.Phaseshifter(phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=4 * np.pi / 5)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 6, phi=3 * np.pi / 2)

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [
            0.0,
            0.0,
            0.0625,
            0.1875,
            0.0,
            0.0625,
            0.1875,
            0.04308374,
            0.02366748,
            0.43324878,
        ],
    )


def test_interferometer():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 1, 1)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 1, 1)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 0, 2)) * np.sqrt(1 / 8)

    T = np.array(
        [
            [0.5, 0.53033009 + 0.53033009j, 0.21650635 + 0.375j],
            [-0.61237244 + 0.61237244j, 0.4330127, 0.24148146 + 0.06470476j],
            [0, -0.48296291 + 0.12940952j, 0.8660254],
        ]
    )

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q(0, 1, 2) | pq.Interferometer(matrix=T)

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [
            0.0,
            0.046875,
            0.015625,
            0.1875,
            0.17307537,
            0.11538358,
            0.32090931,
            0.0192306,
            0.10696977,
            0.01443139,
        ],
    )


def test_kerr():
    xi = np.pi / 3

    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 3), bra=(0, 0, 3)) * 1
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 3), bra=(0, 1, 2)) * (-1j)
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 2), bra=(0, 0, 3)) * 1j

        pq.Q(2) | pq.Kerr(xi=xi)

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert np.isclose(state.norm, 1)

    assert len(nonzero_elements) == 3

    assert np.isclose(nonzero_elements[0][0], 1j * np.exp(1j * xi))
    assert nonzero_elements[0][1] == ((0, 1, 2), (0, 0, 3))

    assert np.isclose(nonzero_elements[1][0], -1j * np.exp(-1j * xi))
    assert nonzero_elements[1][1] == ((0, 0, 3), (0, 1, 2))

    assert np.isclose(nonzero_elements[2][0], 1)
    assert nonzero_elements[2][1] == ((0, 0, 3), (0, 0, 3))


def test_cross_kerr():
    xi = np.pi / 3

    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 3), bra=(0, 0, 3)) * 1
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 3), bra=(0, 1, 2)) * (-1j)
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 2), bra=(0, 0, 3)) * 1j

        pq.Q(1, 2) | pq.CrossKerr(xi=xi)

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert np.isclose(state.norm, 1)

    assert len(nonzero_elements) == 3

    assert np.isclose(nonzero_elements[0][0], -1j * np.exp(-1j * xi))
    assert nonzero_elements[0][1] == ((0, 1, 2), (0, 0, 3))

    assert np.isclose(nonzero_elements[1][0], 1j * np.exp(1j * xi))
    assert nonzero_elements[1][1] == ((0, 0, 3), (0, 1, 2))

    assert np.isclose(nonzero_elements[2][0], 1)
    assert nonzero_elements[2][1] == ((0, 0, 3), (0, 0, 3))


def test_cubic_phase():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.CubicPhase(gamma=0.1)

    simulator = pq.FockSimulator(d=1, config=pq.Config(cutoff=5))
    state = simulator.execute(program).state

    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 25



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import re

import piquasso as pq


def test_correctly_defined_program_executes_without_exception(
    FakeSimulator,
    FakePreparation,
    FakeGate,
    FakeMeasurement,
):
    with pq.Program() as program:
        pq.Q() | FakePreparation()
        pq.Q() | FakeGate()
        pq.Q() | FakeMeasurement()

    simulator = FakeSimulator(d=1)

    simulator.validate(program)
    simulator.execute(program)


def test_program_defined_without_measurement_executes_without_exception(
    FakeSimulator,
    FakePreparation,
    FakeGate,
):
    with pq.Program() as program:
        pq.Q() | FakePreparation()
        pq.Q() | FakeGate()

    simulator = FakeSimulator(d=1)

    simulator.validate(program)
    simulator.execute(program)


def test_program_defined_with_multiple_gates_or_preparations_executes_without_exception(
    FakeSimulator,
    FakePreparation,
    FakeGate,
    FakeMeasurement,
):
    with pq.Program() as program:
        pq.Q() | FakePreparation()
        pq.Q() | FakePreparation()
        pq.Q() | FakePreparation()
        pq.Q() | FakeGate()
        pq.Q() | FakeGate()
        pq.Q() | FakeGate()
        pq.Q() | FakeMeasurement()

    simulator = FakeSimulator(d=1)

    simulator.validate(program)
    simulator.execute(program)


def test_empty_program_executes_without_exception(
    FakeSimulator,
):
    with pq.Program() as program:
        pass

    simulator = FakeSimulator(d=1)

    simulator.validate(program)
    simulator.execute(program)


def test_program_execution_with_unregistered_instruction_raises_InvalidSimulation(
    FakeSimulator,
    FakeGate,
    FakeMeasurement,
):
    class ImproperInstruction(pq.Instruction):
        pass

    with pq.Program() as program:
        pq.Q() | ImproperInstruction()
        pq.Q() | FakeGate()
        pq.Q() | FakeMeasurement()

    simulator = FakeSimulator(d=1)

    with pytest.raises(pq.api.exceptions.InvalidSimulation) as error:
        simulator.validate(program)

    error_message = error.value.args[0]

    assert "No such instruction implemented for this simulator." in error_message
    assert "instruction=ImproperInstruction(modes=())" in error_message

    with pytest.raises(pq.api.exceptions.InvalidSimulation) as error:
        simulator.execute(program)

    error_message = error.value.args[0]

    assert "No such instruction implemented for this simulator." in error_message
    assert "instruction=ImproperInstruction(modes=())" in error_message


def test_program_execution_with_wrong_instruction_order_raises_InvalidSimulation(
    FakeSimulator,
    FakePreparation,
    FakeGate,
    FakeMeasurement,
):
    with pq.Program() as program:
        pq.Q() | FakeGate()
        pq.Q() | FakePreparation()
        pq.Q() | FakeMeasurement()

    simulator = FakeSimulator(d=1)

    with pytest.raises(pq.api.exceptions.InvalidSimulation):
        simulator.validate(program)

    with pytest.raises(pq.api.exceptions.InvalidSimulation):
        simulator.execute(program)


def test_program_execution_with_duplicated_blocks_raises_InvalidSimulation(
    FakeSimulator,
    FakePreparation,
    FakeGate,
    FakeMeasurement,
):
    with pq.Program() as program:
        pq.Q() | FakePreparation()
        pq.Q() | FakeGate()
        pq.Q() | FakeMeasurement()

        pq.Q() | FakePreparation()
        pq.Q() | FakeGate()
        pq.Q() | FakeMeasurement()

    simulator = FakeSimulator(d=1)

    with pytest.raises(pq.api.exceptions.InvalidSimulation):
        simulator.validate(program)

    with pytest.raises(pq.api.exceptions.InvalidSimulation):
        simulator.execute(program)


def test_program_execution_with_multiple_measurements_raises_InvalidSimulation(
    FakeSimulator,
    FakePreparation,
    FakeGate,
    FakeMeasurement,
):
    with pq.Program() as program:
        pq.Q() | FakeGate()
        pq.Q() | FakePreparation()
        pq.Q() | FakeMeasurement()
        pq.Q() | FakeMeasurement()

    simulator = FakeSimulator(d=1)

    with pytest.raises(pq.api.exceptions.InvalidSimulation):
        simulator.validate(program)

    with pytest.raises(pq.api.exceptions.InvalidSimulation):
        simulator.execute(program)


def test_program_execution_with_measurement_not_being_last_raises_InvalidSimulation(
    FakeSimulator,
    FakePreparation,
    FakeGate,
    FakeMeasurement,
):
    with pq.Program() as program:
        pq.Q() | FakeGate()
        pq.Q() | FakeMeasurement()
        pq.Q() | FakePreparation()

    simulator = FakeSimulator(d=1)

    with pytest.raises(pq.api.exceptions.InvalidSimulation):
        simulator.validate(program)

    with pytest.raises(pq.api.exceptions.InvalidSimulation):
        simulator.execute(program)


def test_program_execution_with_initial_state(
    FakeSimulator,
    FakeGate,
    FakeState,
):
    simulator = FakeSimulator(d=1)

    with pq.Program() as program:
        pass

    initial_state = simulator.execute(program).state

    with pq.Program() as program:
        pq.Q() | FakeGate()

    simulator.validate(program)
    simulator.execute(program, initial_state=initial_state)


def test_program_execution_with_initial_state_of_wrong_type_raises_InvalidState(
    FakeSimulator,
    FakeState,
):
    simulator = FakeSimulator(d=1)

    with pq.Program() as program:
        pass

    simulator.validate(program)

    initial_state = simulator.execute(program).state

    class OtherFakeState(FakeState):
        pass

    class OtherFakeSimulator(FakeSimulator):
        _state_class = OtherFakeState

    other_simulator = OtherFakeSimulator(d=1)

    other_simulator.validate(program)

    with pytest.raises(pq.api.exceptions.InvalidState) as exc:
        other_simulator.execute(program, initial_state=initial_state)

    assert "Initial state is specified with type" in exc.value.args[0]


def test_program_execution_with_initial_state_of_wrong_no_of_modes_raises_InvalidState(
    FakeSimulator,
):
    single_mode_simulator = FakeSimulator(d=1)

    with pq.Program() as program:
        pass

    single_mode_simulator.validate(program)

    single_mode_initial_state = single_mode_simulator.execute(program).state

    two_mode_simulator = FakeSimulator(d=2)

    two_mode_simulator.validate(program)

    expected_error_message = (
        "Mismatch in number of specified modes: According to the simulator, "
        "the number of modes should be '2', but the specified 'initial_state' "
        "is defined on '1' modes."
    )

    with pytest.raises(pq.api.exceptions.InvalidState, match=expected_error_message):
        two_mode_simulator.execute(program, initial_state=single_mode_initial_state)


def test_program_execution_with_invalid_modes_raises_InvalidModes_with_single_mode(
    FakeSimulator,
    FakeGate,
):
    with pq.Program() as program:
        pq.Q(0, 1) | FakeGate()

    simulator = FakeSimulator(d=1)

    expected_error_message = re.escape(
        "Instruction 'FakeGate(modes=(0, 1))' addresses mode '1', "
        "which is out of range for the simulator defined on '1' modes. "
        "For a single-mode system, the only valid mode index is '0'."
    )

    with pytest.raises(pq.api.exceptions.InvalidModes, match=expected_error_message):
        simulator.validate(program)

    with pytest.raises(pq.api.exceptions.InvalidModes, match=expected_error_message):
        simulator.execute(program)


def test_program_execution_with_invalid_modes_raises_InvalidModes_with_multiple_modes(
    FakeSimulator,
    FakeGate,
):
    with pq.Program() as program:
        pq.Q(0, 2) | FakeGate()

    simulator = FakeSimulator(d=2)

    expected_error_message = re.escape(
        "Instruction 'FakeGate(modes=(0, 2))' addresses mode '2', "
        "which is out of range for the simulator defined on '2' modes. "
        "Valid mode indices are between '0' and '1' (inclusive)."
    )

    with pytest.raises(pq.api.exceptions.InvalidModes, match=expected_error_message):
        simulator.validate(program)

    with pytest.raises(pq.api.exceptions.InvalidModes, match=expected_error_message):
        simulator.execute(program)


def test_Config_override(FakeSimulator, FakeConfig):
    with pq.Program() as program:
        pass

    simulator = FakeSimulator(d=1)
    state = simulator.execute(program).state

    assert isinstance(simulator.config, FakeConfig)
    assert isinstance(state._config, FakeConfig)


def test_Simulator_repr(FakeSimulator):
    assert (
        repr(FakeSimulator(d=3))
        == "FakeSimulator(d=3, config=Config(), connector=FakeConnector())"
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import piquasso as pq


def code_is_executable(code):
    exec(code)


def test_empty_config_code_generation():
    # This is tested separately as we can not test it using the public interface of
    # piquasso
    config = pq.Config()
    code = config._as_code()
    code_is_executable(code)

    assert code == "pq.Config()"


def test_empty_code_generation():
    with pq.Program() as program:
        pass

    simulator = pq.GaussianSimulator(d=2)

    code = pq.as_code(program, simulator)
    code_is_executable(code)

    assert (
        code
        == f"""\
import numpy as np
import piquasso as pq


with pq.Program() as program:
    pass

simulator = pq.{pq.GaussianSimulator.__name__}(d=2)

result = simulator.execute(program, shots=1)
"""
    )


def test_complicated_code_generation():
    with pq.Program() as program:
        pq.Q(0) | pq.Squeezing(r=0.5, phi=0)
        pq.Q(1) | pq.Squeezing(r=0.5, phi=0)
        pq.Q(0, 1) | pq.Beamsplitter(theta=1, phi=0.7853981633974483)

        pq.Q(0) | pq.Fourier()
        pq.Q(1) | pq.Fourier()

    simulator = pq.GaussianSimulator(d=2, config=pq.Config(seed_sequence=1, cutoff=7))

    code = pq.as_code(program, simulator, shots=10)
    code_is_executable(code)

    assert (
        code
        == f"""\
import numpy as np
import piquasso as pq


with pq.Program() as program:
    pq.Q(0) | pq.Squeezing(r=0.5, phi=0)
    pq.Q(1) | pq.Squeezing(r=0.5, phi=0)
    pq.Q(0, 1) | pq.Beamsplitter(theta=1, phi=0.7853981633974483)
    pq.Q(0) | pq.Fourier()
    pq.Q(1) | pq.Fourier()

simulator = pq.{pq.GaussianSimulator.__name__}(
    d=2, config=pq.Config(seed_sequence=1, cutoff=7)
)

result = simulator.execute(program, shots=10)
"""
    )


def test_numpy_ndarray_parameter_code_generation():
    with pq.Program() as program:
        pq.Q(0, 1) | pq.GeneraldyneMeasurement(
            detection_covariance=np.array(
                [
                    [1.0, 0.0],
                    [0.0, 1.0],
                ],
            ),
        )

    simulator = pq.GaussianSimulator(d=2)

    code = pq.as_code(program, simulator, shots=420)
    code_is_executable(code)

    assert (
        code
        == f"""\
import numpy as np
import piquasso as pq


with pq.Program() as program:
    pq.Q(0, 1) | pq.GeneraldyneMeasurement(detection_covariance=np.array([[1., 0.],
       [0., 1.]]))

simulator = pq.{pq.GaussianSimulator.__name__}(d=2)

result = simulator.execute(program, shots=420)
"""
    )


def test_full_config_code_generation():
    with pq.Program() as program:
        pq.Q(0) | pq.Fourier()
        pq.Q(0, 1) | pq.Beamsplitter(theta=1, phi=0)
        pq.Q(0, 2) | pq.Squeezing2(r=0.5, phi=0)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(
        d=3,
        config=pq.Config(
            seed_sequence=0,
            cache_size=64,
            hbar=2.5,
            use_torontonian=True,
            cutoff=6,
            measurement_cutoff=4,
            dtype=np.float32,
        ),
    )

    code = pq.as_code(program, simulator, shots=100)
    code_is_executable(code)

    assert (
        code
        == f"""\
import numpy as np
import piquasso as pq


with pq.Program() as program:
    pq.Q(0) | pq.Fourier()
    pq.Q(0, 1) | pq.Beamsplitter(theta=1, phi=0)
    pq.Q(0, 2) | pq.Squeezing2(r=0.5, phi=0)
    pq.Q() | pq.ParticleNumberMeasurement()

simulator = pq.{pq.GaussianSimulator.__name__}(
    d=3, config=pq.Config(seed_sequence=0, cache_size=64, hbar=2.5, \
use_torontonian=True, cutoff=6, measurement_cutoff=4, dtype=np.float32)
)

result = simulator.execute(program, shots=100)
"""
    )


def test_Kerr_scalar_parameter_on_multimode_code_generation():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2)) * np.sqrt(1 / 8)

        pq.Q(0) | pq.Kerr(xi=0.2)
        pq.Q(1) | pq.Kerr(xi=0.2)

    simulator = pq.FockSimulator(d=2)

    code = pq.as_code(program, simulator, shots=10)
    code_is_executable(code)

    assert (
        code
        == f"""\
import numpy as np
import piquasso as pq


with pq.Program() as program:
    pq.Q() | pq.Vacuum()
    pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1), coefficient=0.25)
    pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2), coefficient=0.25)
    pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0), coefficient=0.5)
    pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0), coefficient=0.3535533905932738)
    pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2), coefficient=0.3535533905932738)
    pq.Q(0) | pq.Kerr(xi=0.2)
    pq.Q(1) | pq.Kerr(xi=0.2)

simulator = pq.{pq.FockSimulator.__name__}(d=2)

result = simulator.execute(program, shots=10)
"""
    )


def test_Kerr_vector_parameter_code_generation():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2)) * np.sqrt(1 / 8)

        pq.Q(0) | pq.Kerr(xi=-0.1)
        pq.Q(1) | pq.Kerr(xi=0.2)

    simulator = pq.FockSimulator(d=2)

    code = pq.as_code(program, simulator, shots=10)
    code_is_executable(code)

    assert (
        code
        == f"""\
import numpy as np
import piquasso as pq


with pq.Program() as program:
    pq.Q() | pq.Vacuum()
    pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1), coefficient=0.25)
    pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2), coefficient=0.25)
    pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0), coefficient=0.5)
    pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0), coefficient=0.3535533905932738)
    pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2), coefficient=0.3535533905932738)
    pq.Q(0) | pq.Kerr(xi=-0.1)
    pq.Q(1) | pq.Kerr(xi=0.2)

simulator = pq.{pq.FockSimulator.__name__}(d=2)

result = simulator.execute(program, shots=10)
"""
    )


def test_CubicPhase_vector_parameter_code_generation():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2)) * np.sqrt(1 / 8)

        pq.Q(0) | pq.CubicPhase(gamma=-0.1)
        pq.Q(1) | pq.CubicPhase(gamma=0.2)

    simulator = pq.FockSimulator(d=2)

    code = pq.as_code(program, simulator, shots=10)
    code_is_executable(code)

    assert (
        code
        == f"""\
import numpy as np
import piquasso as pq


with pq.Program() as program:
    pq.Q() | pq.Vacuum()
    pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1), coefficient=0.25)
    pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2), coefficient=0.25)
    pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0), coefficient=0.5)
    pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0), coefficient=0.3535533905932738)
    pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2), coefficient=0.3535533905932738)
    pq.Q(0) | pq.CubicPhase(gamma=-0.1)
    pq.Q(1) | pq.CubicPhase(gamma=0.2)

simulator = pq.{pq.FockSimulator.__name__}(d=2)

result = simulator.execute(program, shots=10)
"""
    )


def test_CubicPhase_scalar_parameter_on_multimode_code_generation():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2)) * np.sqrt(1 / 8)

        pq.Q(0) | pq.CubicPhase(gamma=0.2)
        pq.Q(1) | pq.CubicPhase(gamma=0.2)

    simulator = pq.FockSimulator(d=2)

    code = pq.as_code(program, simulator, shots=10)
    code_is_executable(code)

    assert (
        code
        == f"""\
import numpy as np
import piquasso as pq


with pq.Program() as program:
    pq.Q() | pq.Vacuum()
    pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1), coefficient=0.25)
    pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2), coefficient=0.25)
    pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0), coefficient=0.5)
    pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0), coefficient=0.3535533905932738)
    pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2), coefficient=0.3535533905932738)
    pq.Q(0) | pq.CubicPhase(gamma=0.2)
    pq.Q(1) | pq.CubicPhase(gamma=0.2)

simulator = pq.{pq.FockSimulator.__name__}(d=2)

result = simulator.execute(program, shots=10)
"""
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np
import piquasso as pq

from pytest_lazy_fixtures import lf


@pytest.fixture
def tf_function(tf):
    return tf.function


@pytest.fixture
def tf_function_jit(tf):
    return tf.function(jit_compile=True)


def noop_decorator(func):
    return func


def test_tf_function_cvnn_layer_1_mode_1_layers(tf):
    d = 1
    cutoff = 3

    weights = tf.Variable(
        [[0.20961794, -0.00454663, 0.17257116, -0.00007423, -0.12339027, -0.01005965]]
    )

    @tf.function
    def func(weights):
        simulator = pq.PureFockSimulator(
            d=d,
            config=pq.Config(cutoff=cutoff),
            connector=pq.TensorflowConnector(),
        )

        with tf.GradientTape() as tape:
            program = pq.cvqnn.create_program(weights)

            state = simulator.execute(program).state
            mean_position = state.mean_position(0)

        return mean_position, tape.gradient(mean_position, weights)

    mean_position, mean_position_grad = func(weights)

    assert np.allclose(
        mean_position,
        -0.00014718642085721634,
    )
    assert np.allclose(
        mean_position_grad,
        [[0.0, 0.00001151, -0.00000038, 1.9829729, -0.00001934, -0.00001949]],
    )


def test_tf_function_cvnn_layer_2_modes_2_layers(tf):
    d = 2
    cutoff = 3

    weights = tf.Variable(
        [
            [
                -0.01496882,
                -0.00769323,
                -0.23639019,
                0.00605814,
                -0.00198169,
                -0.00070044,
                -0.09768252,
                0.10580704,
                0.01289619,
                0.00523546,
                -0.1072419,
                -0.06125172,
                -0.00887781,
                -0.02219815,
            ],
            [
                -0.15894917,
                0.15323096,
                0.02459804,
                0.00319082,
                -0.0013642,
                0.06503896,
                -0.04608497,
                -0.04319411,
                -0.00326043,
                0.00621172,
                0.06301634,
                0.07470028,
                0.01229031,
                0.00481799,
            ],
        ]
    )

    @tf.function
    def func(weights):
        simulator = pq.PureFockSimulator(
            d=d,
            config=pq.Config(cutoff=cutoff),
            connector=pq.TensorflowConnector(),
        )

        with tf.GradientTape() as tape:
            program = pq.cvqnn.create_program(weights)

            state = simulator.execute(program).state
            mean_position = state.mean_position(0)

        return mean_position, tape.gradient(mean_position, weights)

    mean_position, mean_position_grad = func(weights)

    assert np.allclose(
        mean_position,
        0.019824614605943112,
    )
    assert np.allclose(
        mean_position_grad,
        [
            [
                0.0,
                0.0,
                0.0,
                -0.00177408,
                0.00011175,
                0.00000129,
                0.00000001,
                -0.00009018,
                1.96946914,
                0.17731823,
                0.00313927,
                0.00034899,
                0.00305477,
                0.00035107,
            ],
            [
                -0.00771081,
                -0.00041313,
                0.00346222,
                -0.02683549,
                0.00044953,
                -0.00797526,
                0.00006062,
                0.00336663,
                1.99442205,
                -0.00004225,
                0.00047722,
                -0.00000408,
                0.00381727,
                -0.00000682,
            ],
        ],
    )


def test_tf_function_cvnn_layer_1_mode_1_layers_decorate_with_tf_function(tf):
    d = 1
    cutoff = 3

    weights = tf.Variable(
        [[0.20961794, -0.00454663, 0.17257116, -0.00007423, -0.12339027, -0.01005965]]
    )

    @tf.function
    def func(weights):
        simulator = pq.PureFockSimulator(
            d=d,
            config=pq.Config(cutoff=cutoff),
            connector=pq.TensorflowConnector(decorate_with=tf.function),
        )

        with tf.GradientTape() as tape:
            program = pq.cvqnn.create_program(weights)

            state = simulator.execute(program).state
            mean_position = state.mean_position(0)

        return mean_position, tape.gradient(mean_position, weights)

    mean_position, mean_position_grad = func(weights)

    assert np.allclose(
        mean_position,
        -0.00014718642085721634,
    )
    assert np.allclose(
        mean_position_grad,
        [[0.0, 0.00001151, -0.00000038, 1.9829729, -0.00001934, -0.00001949]],
    )


def test_tf_function_cvnn_layer_2_modes_2_layers_decorate_with_tf_function(tf):
    d = 2
    cutoff = 3

    weights = tf.Variable(
        [
            [
                -0.01496882,
                -0.00769323,
                -0.23639019,
                0.00605814,
                -0.00198169,
                -0.00070044,
                -0.09768252,
                0.10580704,
                0.01289619,
                0.00523546,
                -0.1072419,
                -0.06125172,
                -0.00887781,
                -0.02219815,
            ],
            [
                -0.15894917,
                0.15323096,
                0.02459804,
                0.00319082,
                -0.0013642,
                0.06503896,
                -0.04608497,
                -0.04319411,
                -0.00326043,
                0.00621172,
                0.06301634,
                0.07470028,
                0.01229031,
                0.00481799,
            ],
        ]
    )

    @tf.function
    def func(weights):
        simulator = pq.PureFockSimulator(
            d=d,
            config=pq.Config(cutoff=cutoff),
            connector=pq.TensorflowConnector(decorate_with=tf.function),
        )

        with tf.GradientTape() as tape:
            program = pq.cvqnn.create_program(weights)

            state = simulator.execute(program).state
            mean_position = state.mean_position(0)

        return mean_position, tape.gradient(mean_position, weights)

    mean_position, mean_position_grad = func(weights)

    assert np.allclose(
        mean_position,
        0.019824614605943112,
    )
    assert np.allclose(
        mean_position_grad,
        [
            [
                0.0,
                0.0,
                0.0,
                -0.00177408,
                0.00011175,
                0.00000129,
                0.00000001,
                -0.00009018,
                1.96946914,
                0.17731823,
                0.00313927,
                0.00034899,
                0.00305477,
                0.00035107,
            ],
            [
                -0.00771081,
                -0.00041313,
                0.00346222,
                -0.02683549,
                0.00044953,
                -0.00797526,
                0.00006062,
                0.00336663,
                1.99442205,
                -0.00004225,
                0.00047722,
                -0.00000408,
                0.00381727,
                -0.00000682,
            ],
        ],
    )


def test_tf_function_cvnn_layer_1_mode_1_layers_jit_compile(tf):
    d = 1
    cutoff = 3

    weights = tf.Variable(
        [[0.20961794, -0.00454663, 0.17257116, -0.00007423, -0.12339027, -0.01005965]]
    )

    @tf.function(jit_compile=True)
    def func(weights):
        simulator = pq.PureFockSimulator(
            d=d,
            config=pq.Config(cutoff=cutoff),
            connector=pq.TensorflowConnector(
                decorate_with=tf.function(jit_compile=True)
            ),
        )

        with tf.GradientTape() as tape:
            program = pq.cvqnn.create_program(weights)

            state = simulator.execute(program).state
            mean_position = state.mean_position(0)

        return mean_position, tape.gradient(mean_position, weights)

    mean_position, mean_position_grad = func(weights)

    assert np.allclose(
        mean_position,
        -0.00014718642085721634,
    )
    assert np.allclose(
        mean_position_grad,
        [[0.0, 0.00001151, -0.00000038, 1.9829729, -0.00001934, -0.00001949]],
    )


def test_tf_function_cvnn_layer_2_modes_2_layers_jit_compile(tf):
    d = 2
    cutoff = 3

    weights = tf.Variable(
        [
            [
                -0.01496882,
                -0.00769323,
                -0.23639019,
                0.00605814,
                -0.00198169,
                -0.00070044,
                -0.09768252,
                0.10580704,
                0.01289619,
                0.00523546,
                -0.1072419,
                -0.06125172,
                -0.00887781,
                -0.02219815,
            ],
            [
                -0.15894917,
                0.15323096,
                0.02459804,
                0.00319082,
                -0.0013642,
                0.06503896,
                -0.04608497,
                -0.04319411,
                -0.00326043,
                0.00621172,
                0.06301634,
                0.07470028,
                0.01229031,
                0.00481799,
            ],
        ]
    )

    @tf.function(jit_compile=True)
    def func(weights):
        simulator = pq.PureFockSimulator(
            d=d,
            config=pq.Config(cutoff=cutoff),
            connector=pq.TensorflowConnector(
                decorate_with=tf.function(jit_compile=True)
            ),
        )

        with tf.GradientTape() as tape:
            program = pq.cvqnn.create_program(weights)

            state = simulator.execute(program).state
            mean_position = state.mean_position(0)

        return mean_position, tape.gradient(mean_position, weights)

    mean_position, mean_position_grad = func(weights)

    assert np.allclose(
        mean_position,
        0.019824614605943112,
    )
    assert np.allclose(
        mean_position_grad,
        [
            [
                0.0,
                0.0,
                0.0,
                -0.00177408,
                0.00011175,
                0.00000129,
                0.00000001,
                -0.00009018,
                1.96946914,
                0.17731823,
                0.00313927,
                0.00034899,
                0.00305477,
                0.00035107,
            ],
            [
                -0.00771081,
                -0.00041313,
                0.00346222,
                -0.02683549,
                0.00044953,
                -0.00797526,
                0.00006062,
                0.00336663,
                1.99442205,
                -0.00004225,
                0.00047722,
                -0.00000408,
                0.00381727,
                -0.00000682,
            ],
        ],
    )


@pytest.mark.parametrize(
    "decorator", (noop_decorator, lf("tf_function"), lf("tf_function_jit"))
)
def test_tf_function_cvnn_layer_1_mode_1_layers_custom_initial_state(decorator, tf):
    d = 1
    cutoff = 3

    weights = tf.Variable(
        [[0.20961794, -0.00454663, 0.17257116, -0.00007423, -0.12339027, -0.01005965]],
        dtype=np.float64,
    )

    initial_state_vector = tf.Variable([0.1, 0.3, 0.6], dtype=np.complex128)

    @decorator
    def func(initial_state_vector, weights):
        connector = pq.TensorflowConnector()
        config = pq.Config(cutoff=cutoff)

        initial_state = pq.PureFockState(d=1, connector=connector, config=config)

        initial_state.state_vector = initial_state_vector

        simulator = pq.PureFockSimulator(
            d=d,
            config=config,
            connector=connector,
        )

        with tf.GradientTape() as tape:
            program = pq.cvqnn.create_layers(weights)

            state = simulator.execute(program, initial_state=initial_state).state
            mean_position = state.mean_position(0)

        return mean_position, tape.gradient(mean_position, weights)

    mean_position, mean_position_grad = func(initial_state_vector, weights)

    assert np.allclose(
        mean_position,
        0.28788324741568105,
    )
    assert np.allclose(
        mean_position_grad,
        [[-1.37846195, 0.13790631, -1.38045047, 0.27905517, 0.00014816, -2.28649567]],
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


def test_initial_state_raises_InvalidState_for_nonnormalized_input_state():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0]) * 0.5

    simulator = pq.SamplingSimulator(d=5)
    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        state.validate()

    assert error.value.args[0] == "The state is not normalized: norm=0.25"


def test_initial_state_for_nonnormalized_input_state_validate_False():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0]) * 0.5

    simulator = pq.SamplingSimulator(d=5, config=pq.Config(validate=False))
    state = simulator.execute(program).state

    state.validate()


def test_initial_state_raises_InvalidState_for_occupation_numbers_of_differing_length():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0]) * 1 / np.sqrt(2)
        pq.Q() | pq.StateVector([1, 1, 1]) * 1 / np.sqrt(2)

    simulator = pq.SamplingSimulator(d=5)

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        simulator.execute(program).state

    assert error.value.args[0] == (
        "The occupation numbers '(1, 1, 1)' are not well-defined on '5' modes: "
        "instruction=StateVector(occupation_numbers=(1, 1, 1), "
        "coefficient=0.7071067811865475, modes=(0, 1, 2, 3, 4))"
    )


def test_interferometer_init():
    with pq.Program() as program:
        pass

    simulator = pq.SamplingSimulator(d=5)

    state = simulator.execute(program).state

    expected_interferometer = np.diag(np.ones(state.d, dtype=complex))
    assert np.allclose(state.interferometer, expected_interferometer)


def test_multiple_interferometer_on_neighbouring_modes():
    U = np.array(
        [
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9],
        ],
    )

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])

        pq.Q(0, 1, 2) | pq.Interferometer(U)

    simulator = pq.SamplingSimulator(d=5)
    state = simulator.execute(program).state

    expected_interferometer = np.array(
        [
            [1, 2, 3, 0, 0],
            [4, 5, 6, 0, 0],
            [7, 8, 9, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
        ],
        dtype=complex,
    )

    assert np.allclose(state.interferometer, expected_interferometer)


def test_multiple_interferometer_on_gaped_modes():
    U = np.array(
        [
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9],
        ],
        dtype=complex,
    )

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])

        pq.Q(0, 1, 4) | pq.Interferometer(U)

    simulator = pq.SamplingSimulator(d=5)
    state = simulator.execute(program).state

    expected_interferometer = np.array(
        [
            [1, 2, 0, 0, 3],
            [4, 5, 0, 0, 6],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
            [7, 8, 0, 0, 9],
        ],
        dtype=complex,
    )

    assert np.allclose(state.interferometer, expected_interferometer)


def test_multiple_interferometer_on_reversed_gaped_modes():
    U = np.array(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        dtype=complex,
    )

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])

        pq.Q(4, 3, 1) | pq.Interferometer(U)

    simulator = pq.SamplingSimulator(d=5)
    state = simulator.execute(program).state

    expected_interferometer = np.array(
        [
            [1, 0, 0, 0, 0],
            [0, 9, 0, 8, 7],
            [0, 0, 1, 0, 0],
            [0, 6, 0, 5, 4],
            [0, 3, 0, 2, 1],
        ],
        dtype=complex,
    )

    assert np.allclose(state.interferometer, expected_interferometer)


def test_probability_distribution():
    U = np.array(
        [
            [1, 0, 0],
            [0, -0.54687158 + 0.07993182j, 0.32028583 - 0.76938896j],
            [0, 0.78696803 + 0.27426941j, 0.42419041 - 0.35428818j],
        ],
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([1, 1, 0])
        pq.Q(all) | pq.Interferometer(U)

    simulator = pq.SamplingSimulator(d=3, config=pq.Config(cutoff=3))
    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.30545762, 0.69454239, 0.0, 0.0, 0.0],
    )


def test_get_particle_detection_probability():
    U = np.array(
        [
            [1, 0, 0],
            [0, -0.54687158 + 0.07993182j, 0.32028583 - 0.76938896j],
            [0, 0.78696803 + 0.27426941j, 0.42419041 - 0.35428818j],
        ],
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([1, 1, 0])
        pq.Q(all) | pq.Interferometer(U)

    simulator = pq.SamplingSimulator(d=3)
    state = simulator.execute(program).state

    probability = state.get_particle_detection_probability(occupation_number=(1, 1, 0))

    assert np.allclose(probability, 0.30545762086020883)


def test_get_particle_detection_probability_complex_scenario():
    U = np.array(
        [
            [
                -0.25022099 + 0.32110177j,
                -0.69426529 - 0.49960543j,
                -0.28233272 + 0.15153042j,
            ],
            [
                -0.69028768 + 0.23351228j,
                0.14839865 + 0.49185272j,
                -0.43153658 - 0.13714903j,
            ],
            [
                0.41073351 + 0.36681879j,
                -0.06655274 + 0.00442722j,
                -0.22146243 - 0.80202711j,
            ],
        ],
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([0, 1, 2])
        pq.Q(all) | pq.Interferometer(U)

    simulator = pq.SamplingSimulator(d=3)
    state = simulator.execute(program).state

    probability = state.get_particle_detection_probability(occupation_number=(2, 1, 0))

    assert np.allclose(probability, 0.038483056956364094)


def test_get_particle_detection_probability_on_different_subspace():
    U = np.array(
        [
            [1, 0, 0],
            [0, -0.54687158 + 0.07993182j, 0.32028583 - 0.76938896j],
            [0, 0.78696803 + 0.27426941j, 0.42419041 - 0.35428818j],
        ],
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([1, 1, 0])
        pq.Q(all) | pq.Interferometer(U)

    simulator = pq.SamplingSimulator(d=3)
    state = simulator.execute(program).state

    different_particle_subspace_occupation_number = (3, 1, 0)

    probability = state.get_particle_detection_probability(
        occupation_number=different_particle_subspace_occupation_number
    )

    assert np.allclose(probability, 0.0)


def test_multiple_StateVector_instructions_state_vector():
    initial_occupation_numbers = np.array(
        [
            [1, 1, 0, 1],
            [2, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 0, 0],
        ]
    )
    coefficients = np.sqrt([0.1, 0.2, 0.3, 0.4])
    d = initial_occupation_numbers.shape[1]

    with pq.Program() as program:
        for idx in range(len(coefficients)):
            pq.Q() | pq.StateVector(
                initial_occupation_numbers[idx], coefficient=coefficients[idx]
            )

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

        pq.Q(1) | pq.Phaseshifter(np.pi / 3)

        pq.Q(1, 2) | pq.Beamsplitter5050()

        pq.Q(0, 3) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 7)

    config = pq.Config(cutoff=4)

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(
        sampling_state.state_vector,
        [
            0.63245553,
            0.0,
            0.0,
            0.0,
            0.0,
            -0.15523032 + 0.15523032j,
            -0.07833273 - 0.13567627j,
            -0.07833273 - 0.13567627j,
            -0.21290502 + 0.07449869j,
            -0.16199069 + 0.04340527j,
            -0.22908942 + 0.06138433j,
            -0.00850609 - 0.11350585j,
            -0.16199069 + 0.04340527j,
            -0.00850609 - 0.11350585j,
            -0.11515274 - 0.0129746j,
            -0.02450333 - 0.0894709j,
            0.03791189 - 0.08560374j,
            0.03791189 - 0.08560374j,
            0.00074147 - 0.00404062j,
            0.14143668 - 0.02975653j,
            0.20002168 - 0.04208209j,
            -0.03276254 + 0.02788141j,
            0.14143668 - 0.02975653j,
            -0.03276254 + 0.02788141j,
            -0.04646917 + 0.06369578j,
            0.0,
            0.0,
            -0.19316271 - 0.04756413j,
            0.0,
            -0.27317333 - 0.06726584j,
            -0.09056537 + 0.02373235j,
            0.0,
            -0.19316271 - 0.04756413j,
            -0.09056537 + 0.02373235j,
            -0.05941314 + 0.0318212j,
        ],
    )


def test_multiple_StateVector_instructions_get_particle_detection_probability():
    initial_occupation_numbers = np.array(
        [
            [1, 1, 0, 1],
            [2, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 0, 0],
        ]
    )
    coefficients = np.sqrt([0.1, 0.2, 0.3, 0.4])
    d = initial_occupation_numbers.shape[1]

    with pq.Program() as program:
        for idx in range(len(coefficients)):
            pq.Q() | pq.StateVector(
                initial_occupation_numbers[idx], coefficient=coefficients[idx]
            )

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)

        pq.Q(1) | pq.Phaseshifter(np.pi / 3)

        pq.Q(1, 2) | pq.Beamsplitter5050()

        pq.Q(0, 3) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 7)

    config = pq.Config(cutoff=4)

    sampling_simulator = pq.SamplingSimulator(d=d, config=config)
    sampling_state = sampling_simulator.execute(program).state
    sampling_state.validate()

    assert np.allclose(
        sampling_state.get_particle_detection_probability((1, 1, 0, 1)),
        0.0018507569323483858,
    )


def test_state_vector_with_fock_amplitude_map_sampling():
    amplitude_map = {(0,): 0.6, (1,): 0.8}

    with pq.Program() as program:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map)

    simulator = pq.SamplingSimulator(d=1, config=pq.Config(cutoff=2))

    state = simulator.execute(program).state

    assert np.allclose(state.state_vector, np.array([0.6, 0.8]))


def test_state_vector_with_fock_amplitude_map_and_coefficient_sampling():
    amplitude_map = {(0,): 0.6, (1,): 0.8}
    coefficient = 1 / np.sqrt(2)

    with pq.Program() as program_with_mul:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map) * coefficient

    with pq.Program() as program_with_param:
        pq.Q() | pq.StateVector(
            fock_amplitude_map=amplitude_map, coefficient=coefficient
        )

    simulator = pq.SamplingSimulator(d=1, config=pq.Config(cutoff=2))

    state_with_mul = simulator.execute(program_with_mul).state
    state_with_param = simulator.execute(program_with_param).state

    expected = coefficient * np.array([0.6, 0.8])

    assert np.allclose(state_with_mul.state_vector, expected)
    assert np.allclose(state_with_param.state_vector, expected)


def test_state_vector_with_fock_amplitude_map_invalid_shape_raises_InvalidState():
    amplitude_map = {(0, 0): 0.6}

    with pq.Program() as program:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map)

    simulator = pq.SamplingSimulator(d=1, config=pq.Config(cutoff=2, validate=True))

    with pytest.raises(pq.api.exceptions.InvalidState):
        simulator.execute(program)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso.instructions.measurements import HomodyneMeasurement
from piquasso.api.result import Result
from piquasso.api.exceptions import NotImplementedCalculation

from piquasso._math.fock import get_fock_space_basis

from ..state import PureFockState

from scipy.special import hermite, factorial
from scipy.optimize import brentq

import numpy as np
import numba as nb

from functools import lru_cache


hermite = lru_cache(maxsize=None)(hermite)


def homodyne_measurement(
    state: PureFockState, instruction: HomodyneMeasurement, shots: int
) -> Result:
    modes = instruction.modes

    phi = instruction.params["phi"]

    if state._config.validate and not np.isclose(phi, 0.0):
        raise NotImplementedCalculation(
            "'HomodyneMeasurement' with nonzero rotation angle is not yet supported."
        )

    reduced_state = state.reduced(modes=modes)

    cutoff = state._config.cutoff
    hbar = state._config.hbar

    sqrt_hbar = np.sqrt(hbar)

    reduced_state.normalize()

    density_matrix = reduced_state.density_matrix

    rng = state._config.rng
    small_d = len(modes)

    mean_positions = np.empty(shape=small_d)

    for idx in range(small_d):
        mean_positions[idx] = state.mean_position(modes[idx]) / sqrt_hbar

    uniforms = rng.uniform(size=(shots, small_d))

    samples = np.empty(shape=(shots, small_d))

    B_matrix = get_B_matrix(cutoff)

    if small_d == 1:
        _homodyne_measurement_one_mode(
            density_matrix,
            shots,
            mean_positions[0],
            B_matrix,
            samples,
            uniforms[:, 0],
        )

    else:
        density_matrix_mode_0 = reduced_state.reduced(modes=(0,)).density_matrix

        spaces = []

        for current_d in range(2, state.d + 1):
            spaces.append(get_fock_space_basis(d=current_d, cutoff=cutoff))

        hermites_size = cutoff * (cutoff + 1) // 2

        hermites = np.empty(shape=hermites_size)

        starting_index = 0

        for idx in range(cutoff):
            hermite_polynomial_size = idx + 1
            hermite_polynomial_coeffs = np.array(hermite(idx))
            hermites[starting_index : starting_index + hermite_polynomial_size] = (
                hermite_polynomial_coeffs
            )
            starting_index += hermite_polynomial_size

        _homodyne_measurement_one_mode(
            density_matrix_mode_0,
            shots=shots,
            mean_position=mean_positions[0],
            B_matrix=B_matrix,
            samples=samples[:, 0],
            uniforms=uniforms[:, 0],
        )

        reduced_density_matrices = []
        for current_d in range(2, reduced_state.d + 1):
            reduced_state_to_current = reduced_state.reduced(
                modes=tuple(range(current_d))
            )
            reduced_density_matrices.append(reduced_state_to_current.density_matrix)

        new_density_matrix = np.empty_like(density_matrix_mode_0)

        _do_sample_homodyne_multimodes(
            shots,
            samples,
            new_density_matrix,
            reduced_density_matrices,
            spaces,
            hermites,
            mean_positions,
            B_matrix,
            uniforms,
        )

    return Result(state=state, samples=sqrt_hbar * samples)


def _do_sample_homodyne_multimodes(
    shots,
    samples,
    new_density_matrix,
    reduced_density_matrices,
    spaces,
    hermites,
    mean_positions,
    B_matrix,
    uniforms,
):
    for i in range(shots):
        _homodyne_measurement_multimode_sample(
            samples[i],
            new_density_matrix,
            reduced_density_matrices,
            spaces,
            hermites,
            mean_positions,
            B_matrix,
            uniforms[i],
        )


def _homodyne_measurement_multimode_sample(
    positions,
    new_density_matrix,
    reduced_density_matrices,
    spaces,
    hermites,
    mean_positions,
    B_matrix,
    uniforms,
):
    d = len(reduced_density_matrices) + 1
    cutoff = new_density_matrix.shape[0]

    for current_d in range(2, d + 1):
        density_matrix = reduced_density_matrices[current_d - 2]

        new_density_matrix.fill(0.0)

        space = spaces[current_d - 2]

        hermite_terms = get_hermite_terms(hermites, positions, space, current_d, cutoff)

        new_density_matrix = _get_density_matrix_on_next_mode(
            new_density_matrix, density_matrix, space, hermite_terms
        )

        _homodyne_measurement_one_mode(
            new_density_matrix,
            shots=1,
            mean_position=mean_positions[current_d - 1],
            B_matrix=B_matrix,
            samples=positions[current_d - 1 :],
            uniforms=uniforms[current_d - 1 :],
        )


def _homodyne_measurement_one_mode(
    density_matrix, shots, mean_position, B_matrix, samples, uniforms
):
    poly = get_integral_poly(density_matrix, B_matrix)

    lower, upper, almost_0, almost_1 = get_interval(mean_position, poly)

    for idx in range(shots):
        inverse_sample = uniforms[idx] * (almost_1 - almost_0) + almost_0

        sample = brentq(
            integral_m_inverse_sample, a=lower, b=upper, args=(poly, inverse_sample)
        )

        samples[idx] = sample


@nb.njit(cache=True)
def _get_density_matrix_on_next_mode(
    new_density_matrix, density_matrix, space, hermite_terms
):
    for row_idx in range(density_matrix.shape[0]):
        row_occupation_number = space[row_idx]

        row_term = hermite_terms[row_idx]

        small_row_idx = row_occupation_number[-1]

        for col_idx in range(row_idx):
            col_occupation_number = space[col_idx]

            col_term = hermite_terms[col_idx]

            small_col_idx = col_occupation_number[-1]

            row_col_term = row_term * col_term

            new_density_matrix[small_row_idx, small_col_idx] += (
                density_matrix[row_idx, col_idx] * row_col_term
            )

            new_density_matrix[small_col_idx, small_row_idx] += (
                density_matrix[col_idx, row_idx] * row_col_term
            )

        diagonal = density_matrix[row_idx, row_idx] * row_term**2

        new_density_matrix[small_row_idx, small_row_idx] += diagonal

    new_density_matrix /= np.trace(new_density_matrix)

    return new_density_matrix


@nb.njit(cache=True)
def polyeval(p, x):
    y = 0.0

    for pv in p:
        y = y * x + pv

    return y


@nb.njit(cache=True)
def eval_poly_term(p, x):
    y = polyeval(p, x)

    return y * np.exp(-(x**2))


@nb.njit(cache=True)
def erf(x):
    """Rational approximation (7.1.25) by Abramowitz and Stegun, see
    https://personal.math.ubc.ca/%7Ecbm/aands/page_299.htm
    """
    if x < 0.0:
        return -erf(-x)

    a3 = 0.7478556
    a2 = -0.0958798
    a1 = 0.3480242
    p = 0.47047

    t = 1 / (1 + p * x)

    y = 1 - (a1 * t + a2 * t**2 + a3 * t**3) * np.exp(-(x**2))

    return y


@nb.njit(cache=True)
def integral(poly, x):
    return (erf(x) + 1) / 2 - eval_poly_term(poly, x)


@nb.njit(cache=True)
def integral_m_inverse_sample(x, poly, inverse_sample):
    return integral(poly, x) - inverse_sample


@nb.njit(cache=True)
def add_poly(poly1, poly2):
    n1 = poly1.shape[0]
    n2 = poly2.shape[0]
    n = max(n1, n2)

    new_poly = np.zeros(n, dtype=poly1.dtype)

    for idx in range(n1):
        addend = poly1[n1 - idx - 1]
        new_poly[n - idx - 1] += addend

    for idx in range(n2):
        addend = poly2[n2 - idx - 1]
        new_poly[n - idx - 1] += addend

    return new_poly


@nb.njit(cache=True)
def get_integral_poly(density_matrix, B_matrix):
    size = density_matrix.shape[0]

    integral_poly = np.array([0.0])

    starting_index = 0

    for row_idx in range(size):
        for col_idx in range(row_idx):
            stopping_index = row_idx + col_idx
            sliced_I = B_matrix[starting_index : starting_index + stopping_index]

            poly_to_add = sliced_I * 2 * np.real(density_matrix[row_idx, col_idx])
            starting_index += stopping_index

            integral_poly = add_poly(poly_to_add, integral_poly)

        stopping_index = max(2 * row_idx, 1)
        sliced_I = B_matrix[starting_index : starting_index + stopping_index]
        integral_poly = add_poly(
            sliced_I * np.real(density_matrix[row_idx, row_idx]),
            integral_poly,
        )
        starting_index += stopping_index

    return integral_poly


@lru_cache(maxsize=None)
def get_B_matrix(cutoff):
    B_matrix = []

    for row_idx in range(cutoff):
        row_list = []
        for col_idx in range(row_idx + 1):
            row_list.extend(np.array(B(row_idx, col_idx)).tolist())

        B_matrix.extend(row_list)

    return np.array(B_matrix)


@nb.njit(cache=True)
def get_interval(mean_position, poly):
    lower = mean_position - 5.0
    upper = mean_position + 5.0

    almost_0 = integral(poly, lower)
    almost_1 = integral(poly, upper)

    lower_tol = 1e-10
    upper_tol = 1.0 - lower_tol

    while almost_0 > lower_tol:
        lower -= 5.0
        almost_0 = integral(poly, lower)

    while almost_1 < upper_tol:
        upper += 5.0
        almost_1 = integral(poly, upper)

    return lower, upper, almost_0, almost_1


@nb.njit(cache=True)
def _get_hermite_vals(hermites, current_d, positions, cutoff):
    hermite_vals = np.empty(shape=(cutoff, current_d - 1))

    starting_index = 0
    for idx in range(cutoff):
        size = idx + 1
        hermite_polynomial_coeffs = hermites[starting_index : starting_index + size]
        starting_index += size
        for jdx in range(current_d - 1):
            hermite_vals[idx, jdx] = polyeval(hermite_polynomial_coeffs, positions[jdx])

    return hermite_vals


@nb.njit(cache=True)
def _do_get_hermite_terms(space, hermite_vals, current_d):
    hermite_terms = np.empty(shape=space.shape[0])

    for idx in range(space.shape[0]):
        occupation_number = space[idx]

        term = 1.0
        for jdx in range(current_d - 1):
            term *= hermite_vals[occupation_number[jdx], jdx]

        hermite_terms[idx] = term

    return hermite_terms


@nb.njit(cache=True)
def get_hermite_terms(hermites, positions, space, current_d, cutoff):
    hermite_vals = _get_hermite_vals(hermites, current_d, positions, cutoff)

    return _do_get_hermite_terms(space, hermite_vals, current_d)


@lru_cache(maxsize=None)
def B(n, m):
    if n > m:
        return B(m, n)

    normalizer = np.sqrt(factorial(n) * factorial(m) * 2 ** (n + m) * np.pi)

    sum_ = np.poly1d([0.0])

    for k in range(n):
        sum_ += hermite(n - k) * hermite(m - k - 1) * 2**k / factorial(n - k)

    sum_ = sum_ * factorial(n) / normalizer

    if n == m:
        return sum_

    return hermite(m - n - 1) * 2**n * factorial(n) / normalizer + sum_



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq
from piquasso.decompositions.clements import (
    get_weights_from_interferometer,
    get_interferometer_from_weights,
)

import numpy as np

import jax.numpy as jnp

from jax import jit, jacrev


def test_Interferometer_2_by_2_gradient():
    connector = pq.JaxConnector()
    theta = np.pi / 5

    @jit
    def calculate_state_vector(theta):
        U = jnp.array(
            [[jnp.cos(theta), -jnp.sin(theta)], [jnp.sin(theta), jnp.cos(theta)]]
        )

        with pq.Program() as program:
            pq.Q(0, 1) | pq.StateVector([0, 1])
            pq.Q(0, 1) | pq.Interferometer(U)

        fock_simulator = pq.fermionic.PureFockSimulator(d=2, connector=connector)

        fock_state = fock_simulator.execute(program).state

        return fock_state.state_vector

    state_vector = calculate_state_vector(np.pi / 5)

    assert np.allclose(state_vector, [0, -np.sin(theta), np.cos(theta), 0])

    calculate_state_vector_jac = jit(jacrev(calculate_state_vector, holomorphic=True))

    # NOTE: `holomorphic=True` requrires both complex inputs and outputs
    state_vector_jac = calculate_state_vector_jac(theta + 0.0j)

    assert np.allclose(state_vector_jac, [0, -np.cos(theta), -np.sin(theta), 0])


@pytest.mark.monkey
def test_Interferometer_3_by_3_random(generate_unitary_matrix):
    connector = pq.JaxConnector()

    d = 3
    U = generate_unitary_matrix(d)

    @jit
    def calculate_state_vector(U):
        with pq.Program() as program:
            # NOTE: This violates the parity superselection rule, but the simulator
            # should permit it.
            pq.Q(0, 1, 2) | pq.StateVector([0, 0, 1]) / np.sqrt(2)
            pq.Q(0, 1, 2) | pq.StateVector([0, 1, 1]) / np.sqrt(2)
            pq.Q(0, 1, 2) | pq.Interferometer(U)

        fock_simulator = pq.fermionic.PureFockSimulator(d=d, connector=connector)

        fock_state = fock_simulator.execute(program).state

        return fock_state.state_vector

    state_vector = calculate_state_vector(U)

    assert np.allclose(
        state_vector,
        1
        / np.sqrt(2)
        * np.array(
            [
                0.0,
                U[0, 2],
                U[1, 2],
                U[2, 2],
                U[0, 1] * U[1, 2] - U[0, 2] * U[1, 1],
                U[0, 1] * U[2, 2] - U[0, 2] * U[2, 1],
                U[1, 1] * U[2, 2] - U[1, 2] * U[2, 1],
                0.0,
            ]
        ),
    )

    calculate_state_vector_jac = jit(jacrev(calculate_state_vector, holomorphic=True))

    state_vector_jac = calculate_state_vector_jac(U)

    def E(i, j):
        # Utility function for basis matrices.
        matrix = np.zeros(shape=(d, d))
        matrix[i, j] = 1.0
        return matrix

    assert np.allclose(state_vector_jac[0], np.zeros(shape=(d, d)))
    assert np.allclose(state_vector_jac[7], np.zeros(shape=(d, d)))

    assert np.allclose(state_vector_jac[1], E(0, 2) / np.sqrt(2))
    assert np.allclose(state_vector_jac[2], E(1, 2) / np.sqrt(2))
    assert np.allclose(state_vector_jac[3], E(2, 2) / np.sqrt(2))

    assert np.allclose(
        state_vector_jac[4],
        (E(0, 1) * U[1, 2] + U[0, 1] * E(1, 2) - E(0, 2) * U[1, 1] - U[0, 2] * E(1, 1))
        / np.sqrt(2),
    )

    assert np.allclose(
        state_vector_jac[5],
        (E(0, 1) * U[2, 2] + U[0, 1] * E(2, 2) - E(0, 2) * U[2, 1] - U[0, 2] * E(2, 1))
        / np.sqrt(2),
    )

    assert np.allclose(
        state_vector_jac[6],
        (E(1, 1) * U[2, 2] + U[1, 1] * E(2, 2) - E(1, 2) * U[2, 1] - U[1, 2] * E(2, 1))
        / np.sqrt(2),
    )


@pytest.mark.monkey
def test_parametrized_circuit_gradient_clements_random(generate_unitary_matrix):
    connector = pq.JaxConnector()

    d = 4

    @jit
    def calculate_state_vector(weights):
        U = get_interferometer_from_weights(weights.real, d, connector, np.complex128)

        with pq.Program() as program:
            pq.Q() | pq.StateVector([0, 0, 1, 1]) / np.sqrt(2)
            pq.Q() | pq.StateVector([1, 1, 0, 0]) / np.sqrt(2)
            pq.Q() | pq.Interferometer(U)

        fock_simulator = pq.fermionic.PureFockSimulator(
            d=d, connector=connector, config=pq.Config(cutoff=d + 1)
        )

        fock_state = fock_simulator.execute(program).state

        return fock_state.state_vector

    U = generate_unitary_matrix(d)

    weights = get_weights_from_interferometer(U, connector)

    state_vector = calculate_state_vector(weights)

    assert np.isclose(
        state_vector.conj() @ state_vector, 1.0
    ), "The resulting state vector should be normalized."

    assert state_vector.shape == (2**d,)

    calculate_state_vector_jac = jit(jacrev(calculate_state_vector, holomorphic=True))

    # NOTE: `holomorphic=True` requrires both complex inputs and outputs
    state_vector_jac = calculate_state_vector_jac(weights + 0.0j)

    assert state_vector_jac.shape == (2**d, len(weights))

    for i in range(2**d):
        np.isclose(
            state_vector.conj() @ state_vector_jac[:, i]
            + state_vector_jac[:, i].conj() @ state_vector,
            0.0,
        ), "Gradient of state vector norm should be 0."



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This code has been copied from the following website with minor modifications:
https://strawberryfields.ai/photonics/demos/run_quantum_neural_network.html
"""

import time
import numpy as np
import tensorflow as tf
import strawberryfields as sf
from strawberryfields import ops


def interferometer(params, q):
    N = len(q)
    theta = params[: N * (N - 1) // 2]
    phi = params[N * (N - 1) // 2 : N * (N - 1)]
    rphi = params[-N + 1 :]

    if N == 1:
        # the interferometer is a single rotation
        ops.Rgate(rphi[0]) | q[0]
        return

    n = 0  # keep track of free parameters

    # Apply the rectangular beamsplitter array
    # The array depth is N
    for i in range(N):
        for j, (q1, q2) in enumerate(zip(q[:-1], q[1:])):
            # skip even or odd pairs depending on layer
            if (i + j) % 2 != 1:
                ops.BSgate(theta[n], phi[n]) | (q1, q2)
                n += 1

    # apply the final local phase shifts to all modes except the last one
    for i in range(max(1, N - 1)):
        ops.Rgate(rphi[i]) | q[i]


def layer(params, q):
    N = len(q)
    M = int(N * (N - 1)) + max(1, N - 1)

    int1 = params[:M]
    s = params[M : M + N]
    int2 = params[M + N : 2 * M + N]
    dr = params[2 * M + N : 2 * M + 2 * N]
    dp = params[2 * M + 2 * N : 2 * M + 3 * N]
    k = params[2 * M + 3 * N : 2 * M + 4 * N]

    # begin layer
    interferometer(int1, q)

    for i in range(N):
        ops.Sgate(s[i]) | q[i]

    interferometer(int2, q)

    for i in range(N):
        ops.Dgate(dr[i], dp[i]) | q[i]
        ops.Kgate(k[i]) | q[i]


def init_weights(modes, layers, active_sd=0.0001, passive_sd=0.1):
    # Number of interferometer parameters:
    M = int(modes * (modes - 1)) + max(1, modes - 1)

    # Create the TensorFlow variables
    int1_weights = tf.random.normal(shape=[layers, M], stddev=passive_sd)
    s_weights = tf.random.normal(shape=[layers, modes], stddev=active_sd)
    int2_weights = tf.random.normal(shape=[layers, M], stddev=passive_sd)
    dr_weights = tf.random.normal(shape=[layers, modes], stddev=active_sd)
    dp_weights = tf.random.normal(shape=[layers, modes], stddev=passive_sd)
    k_weights = tf.random.normal(shape=[layers, modes], stddev=active_sd)

    weights = tf.concat(
        [int1_weights, s_weights, int2_weights, dr_weights, dp_weights, k_weights],
        axis=1,
    )

    weights = tf.Variable(weights)

    return weights


# set the random seed
tf.random.set_seed(137)
np.random.seed(137)


# define width and depth of CV quantum neural network
modes = 7
layers = 1
cutoff = 7

# defining desired state (single photon state)
target_state = np.zeros([cutoff] * modes)
target_state[([1] + [0] * (modes - 1))] = 1.0
target_state = tf.constant(target_state, dtype=tf.complex64)


# initialize engine and program
eng = sf.Engine(backend="tf", backend_options={"cutoff_dim": cutoff})
qnn = sf.Program(modes)

# initialize QNN weights
weights = init_weights(modes, layers)  # our TensorFlow weights
num_params = np.prod(weights.shape)  # total number of parameters in our model


# Create array of Strawberry Fields symbolic gate arguments, matching
# the size of the weights Variable.
sf_params = np.arange(num_params).reshape(weights.shape).astype(str)
sf_params = np.array([qnn.params(*i) for i in sf_params])


# Construct the symbolic Strawberry Fields program by
# looping and applying layers to the program.
with qnn.context as q:
    for i in range(modes):
        sf.ops.Dgate(0.1) | q[i]

    for k in range(layers):
        layer(sf_params[k], q)

with tf.GradientTape() as tape:
    mapping = {
        p.name: w for p, w in zip(sf_params.flatten(), tf.reshape(weights, [-1]))
    }
    start_time = time.time()
    state = eng.run(qnn, args=mapping).state
    print("EXECUTION TIME: ", time.time() - start_time)

    ket = state.ket()

    cost = tf.reduce_sum(tf.abs(ket - target_state))


start_time = time.time()
gradient = tape.gradient(cost, weights)
print("gradient:", gradient)
print("JACOBIAN CALCULATION TIME: ", time.time() - start_time)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso._math.pfaffian import pfaffian
from piquasso.api.connector import BaseConnector

from .connections import (
    calculate_interferometer_on_fermionic_fock_space,
    apply_fermionic_passive_linear_to_state_vector,
)


def instancemethod(func):
    def wrapped(self, *args, **kwargs):
        return func(*args, **kwargs)

    return wrapped


class BuiltinConnector(BaseConnector):
    """Base class for built-in connectors."""

    range = range

    calculate_interferometer_on_fermionic_fock_space = (
        calculate_interferometer_on_fermionic_fock_space
    )

    apply_fermionic_passive_linear_to_state_vector = (
        apply_fermionic_passive_linear_to_state_vector
    )

    def custom_gradient(self, func):
        def noop_grad(*args, **kwargs):
            result, _ = func(*args, **kwargs)
            return result

        return noop_grad

    def accumulator(self, dtype, size, **kwargs):
        return []

    def write_to_accumulator(self, accumulator, index, value):
        accumulator.append(value)

        return accumulator

    def stack_accumulator(self, accumulator):
        return self.forward_pass_np.stack(accumulator)

    def decorator(self, func):
        return func

    def gather_along_axis_1(self, array, indices):
        return array[:, indices]

    def transpose(self, matrix):
        return self.np.transpose(matrix)

    def embed_in_identity(self, matrix, indices, dim):
        embedded_matrix = self.np.identity(dim, dtype=matrix.dtype)

        embedded_matrix = self.assign(embedded_matrix, indices, matrix)

        return embedded_matrix

    def calculate_interferometer_on_fock_space(self, interferometer, helper_indices):
        """
        This implementation uses Eq. (71) from
        https://quantum-journal.org/papers/q-2020-11-30-366/pdf/
        """

        np = self.forward_pass_np
        subspace_representations = []

        subspace_representations.append(np.array([[1.0]], dtype=interferometer.dtype))
        subspace_representations.append(interferometer)

        cutoff = len(helper_indices[0]) + 2

        for n in range(2, cutoff):
            subspace_indices = helper_indices[0][n - 2]
            first_subspace_indices = helper_indices[2][n - 2]

            first_nonzero_indices = helper_indices[1][n - 2]
            sqrt_occupation_numbers = helper_indices[3][n - 2]
            sqrt_first_occupation_numbers = helper_indices[4][n - 2]

            first_part_partially_indexed = interferometer[first_nonzero_indices]
            second = self.gather_along_axis_1(
                subspace_representations[n - 1][first_subspace_indices],
                indices=subspace_indices,
            )

            matrix = np.einsum(
                "ij,kj,kij->ki",
                sqrt_occupation_numbers,
                first_part_partially_indexed,
                second,
            )

            new_subspace_representation = (
                matrix / sqrt_first_occupation_numbers[:, None]
            )

            subspace_representations.append(
                new_subspace_representation.astype(interferometer.dtype)
            )

        return subspace_representations

    def pfaffian(self, matrix):
        """
        Lazy function for calculating the pfaffian.

        Note:
            There are faster algorithms, but this is fine for now.
        """
        return pfaffian(matrix)

    def real_logm(self, matrix):
        """Calculates the real logarithm of a matrix.

        Note:
            This function does not verify the existence of the real logarithm, and it
            will return with a wrong result if such matrix is provided. Note also, that
            this algorithm can certainly be made more efficient, but this is fine for
            now.
        """

        np = self.np

        eigvals, U = np.linalg.eig(matrix)

        I = np.identity(2)
        J = np.array([[0, 1], [-1, 0]])
        WdJW = 1j * np.array([[1, 0], [0, -1]])

        D = np.zeros_like(matrix, dtype=complex)

        forbidden_indices = []

        for index, eigval in enumerate(eigvals):
            if index in forbidden_indices:
                continue

            if np.isclose(np.imag(eigval), 0.0) and eigval >= 0.0:
                D = self.assign(D, (index, index), np.log(eigval))

            else:
                for conjugate_index, conjugate_eigval in enumerate(
                    eigvals[index + 1 :]
                ):
                    if np.isclose(eigval.conj(), conjugate_eigval):
                        break

                forbidden_indices.append(index + conjugate_index + 1)

                indices = np.ix_(*[np.array([index, index + conjugate_index + 1])] * 2)

                r = np.abs(eigval)
                phi = np.angle(eigval)

                if np.isclose(np.imag(eigval), 0.0):
                    D = self.assign(D, indices, np.log(r) * I + phi * J)
                else:
                    D = self.assign(D, indices, np.log(r) * I + phi * WdJW)

        return U @ D @ np.linalg.inv(U)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import piquasso as pq


def test_loads_blackbird_parses_operations():
    blackbird_code = """name StateTeleportation
        version 1.0

        BSgate(0.7853981633974483, 0) | [1, 2]
        Rgate(0.7853981633974483) | 1
        Vgate(0.5) | 1
        """

    program = pq.Program()

    program.loads_blackbird(blackbird_code)

    assert len(program.instructions) == 3

    assert program.instructions[0] == pq.Beamsplitter(
        theta=np.pi / 4, phi=0.0
    ).on_modes(1, 2)
    assert program.instructions[1] == pq.Phaseshifter(phi=np.pi / 4).on_modes(1)
    assert program.instructions[2] == pq.CubicPhase(gamma=0.5).on_modes(1)


def test_loads_blackbird_parses_operations_with_default_arguments():
    blackbird_code = """name StateTeleportation
        version 1.0

        BSgate() | [1, 2]
        Rgate(0.7853981633974483) | 1
        """

    program = pq.Program()

    program.loads_blackbird(blackbird_code)

    assert len(program.instructions) == 2

    assert program.instructions[0] == pq.Beamsplitter(
        theta=np.pi / 4, phi=0.0
    ).on_modes(1, 2)
    assert program.instructions[1] == pq.Phaseshifter(phi=np.pi / 4).on_modes(1)


def test_loads_blackbird_parses_operations_with_classes_registered_separately():
    blackbird_code = """name StateTeleportation
        version 1.0

        BSgate(0.7853981633974483, 0) | [1, 2]
        Rgate(0.7853981633974483) | 1
        """

    class Beamsplitter(pq.Instruction):
        pass

    program = pq.Program()

    program.loads_blackbird(blackbird_code)

    assert len(program.instructions) == 2

    assert program.instructions[0].__class__ is Beamsplitter

    # Teardown
    pq.Instruction.set_subclass(pq.Beamsplitter)
    assert pq.Instruction.get_subclass("Beamsplitter") is pq.Beamsplitter


def test_loads_blackbird_preserves_exising_operations():
    blackbird_code = """name StateTeleportation
        version 1.0

        BSgate(0.7853981633974483, 0) | [1, 2]
        Rgate(0.7853981633974483) | 1
        """

    program = pq.Program()

    squeezing = pq.Squeezing(r=np.log(2), phi=np.pi / 2)

    with program:
        pq.Q(0) | squeezing

    program.loads_blackbird(blackbird_code)

    assert len(program.instructions) == 3

    assert program.instructions[0] == squeezing
    assert program.instructions[1] == pq.Beamsplitter(
        theta=np.pi / 4, phi=0.0
    ).on_modes(1, 2)
    assert program.instructions[2] == pq.Phaseshifter(phi=np.pi / 4).on_modes(1)


def test_loads_blackbird_with_execution(gaussian_state_assets):
    blackbird_code = """name StateTeleportation
        version 1.0

        BSgate(0.7853981633974483, 3.141592653589793) | [1, 2]
        Rgate(0.7853981633974483) | 1
        """

    program = pq.Program()

    simulator = pq.GaussianSimulator(d=3)

    squeezing = pq.Squeezing(r=np.log(2), phi=np.pi / 2)

    with program:
        pq.Q(1) | squeezing

    program.loads_blackbird(blackbird_code)

    state = simulator.execute(program).state

    expected_state = gaussian_state_assets.load()

    assert state == expected_state


def test_load_blackbird_from_file_with_execution(gaussian_state_assets, tmpdir):
    blackbird_code = """name StateTeleportation
        version 1.0

        BSgate(0.7853981633974483, 3.141592653589793) | [1, 2]
        Rgate(0.7853981633974483) | 1
        """

    blackbird_file = tmpdir.join("example-blackbird-code.xbb")

    blackbird_file.write(blackbird_code)

    program = pq.Program()

    simulator = pq.GaussianSimulator(d=3)

    squeezing = pq.Squeezing(r=np.log(2), phi=np.pi / 2)

    with program:
        pq.Q(1) | squeezing

    program.load_blackbird(blackbird_file.strpath)

    state = simulator.execute(program).state

    expected_state = gaussian_state_assets.load()

    assert state == expected_state



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pytest_lazy_fixtures import lf

import numpy as np

import piquasso as pq

from piquasso._math.linalg import is_unitary, is_diagonal
from piquasso._math.symplectic import is_symplectic, xp_symplectic_form
from piquasso._math.decompositions import (
    takagi,
    williamson,
)

from piquasso._simulators.connectors import (
    NumpyConnector,
    JaxConnector,
)


@pytest.mark.parametrize("connector", [NumpyConnector(), lf("tensorflow_connector")])
def test_takagi_on_real_symmetric_2_by_2_matrix(connector):
    matrix = np.array(
        [
            [1, 2],
            [2, 1],
        ],
        dtype=complex,
    )

    singular_values, unitary = takagi(matrix, connector)

    assert is_unitary(unitary)
    assert np.allclose(np.abs(singular_values), singular_values)
    assert np.allclose(matrix, unitary @ np.diag(singular_values) @ unitary.transpose())


@pytest.mark.parametrize("connector", [NumpyConnector(), lf("tensorflow_connector")])
def test_takagi_on_complex_symmetric_2_by_2_matrix_with_multiplicities(connector):
    matrix = np.array(
        [
            [1, 2j],
            [2j, 1],
        ],
        dtype=complex,
    )

    singular_values, unitary = takagi(matrix, connector)

    assert is_unitary(unitary)
    assert np.allclose(np.abs(singular_values), singular_values)
    assert np.allclose(matrix, unitary @ np.diag(singular_values) @ unitary.transpose())


@pytest.mark.parametrize("connector", [NumpyConnector(), lf("tensorflow_connector")])
def test_takagi_on_real_symmetric_3_by_3_matrix(connector):
    matrix = np.array(
        [
            [1, 2, 3],
            [2, 1, 5],
            [3, 5, 9],
        ],
        dtype=complex,
    )

    singular_values, unitary = takagi(matrix, connector)

    assert is_unitary(unitary)
    assert np.allclose(np.abs(singular_values), singular_values)
    assert np.allclose(matrix, unitary @ np.diag(singular_values) @ unitary.transpose())


@pytest.mark.parametrize("connector", [NumpyConnector(), lf("tensorflow_connector")])
def test_takagi_on_complex_symmetric_3_by_3_matrix(connector):
    matrix = np.array(
        [
            [1, 2, 3j],
            [2, 1, 5j],
            [3j, 5j, 9],
        ],
    )

    singular_values, unitary = takagi(matrix, connector)

    assert is_unitary(unitary)
    assert np.allclose(np.abs(singular_values), singular_values)
    assert np.allclose(matrix, unitary @ np.diag(singular_values) @ unitary.transpose())


@pytest.mark.monkey
@pytest.mark.parametrize("connector", [NumpyConnector(), lf("tensorflow_connector")])
def test_takagi_on_complex_symmetric_6_by_6_matrix_with_multiplicities(
    connector,
    generate_unitary_matrix,
):
    singular_values = np.array([1, 1, 2, 2, 2, 3], dtype=complex)

    unitary = generate_unitary_matrix(6)

    matrix = unitary @ np.diag(singular_values) @ unitary.transpose()

    calculated_singular_values, calculated_unitary = takagi(matrix, connector)

    assert is_unitary(calculated_unitary)
    assert np.allclose(np.abs(calculated_singular_values), calculated_singular_values)
    assert np.allclose(
        matrix,
        calculated_unitary
        @ np.diag(calculated_singular_values)
        @ calculated_unitary.transpose(),
    )


@pytest.mark.monkey
@pytest.mark.parametrize("N", [2, 3, 4, 5, 6])
@pytest.mark.parametrize("connector", [NumpyConnector(), lf("tensorflow_connector")])
def test_takagi_on_complex_symmetric_N_by_N_matrix(
    N, connector, generate_complex_symmetric_matrix
):
    matrix = generate_complex_symmetric_matrix(N)
    singular_values, unitary = takagi(matrix, connector)

    assert is_unitary(unitary)
    assert np.allclose(np.abs(singular_values), singular_values)
    assert np.allclose(matrix, unitary @ np.diag(singular_values) @ unitary.transpose())


@pytest.mark.parametrize("connector", [NumpyConnector(), JaxConnector()])
def test_williamson_with_identity(connector):
    covariance_matrix = np.identity(4)
    symplectic, diagonal = williamson(covariance_matrix, connector)

    assert is_diagonal(diagonal)
    assert is_symplectic(symplectic, form_func=xp_symplectic_form)
    assert np.all(np.isreal(symplectic))
    assert np.allclose(covariance_matrix, symplectic @ diagonal @ symplectic.T)


@pytest.mark.parametrize("connector", [NumpyConnector(), JaxConnector()])
def test_williamson_with_diagonal_matrix(connector):
    covariance_matrix = np.diag([1, 2, 3, 4])
    symplectic, diagonal = williamson(covariance_matrix, connector)

    assert is_diagonal(diagonal)
    assert is_symplectic(symplectic, form_func=xp_symplectic_form)
    assert np.all(np.isreal(symplectic))
    assert np.allclose(covariance_matrix, symplectic @ diagonal @ symplectic.T)


@pytest.mark.parametrize("connector", [NumpyConnector(), JaxConnector()])
def test_williamson_with_squeezed_covariance_matrix(connector):
    d = 3
    with pq.Program() as program:
        pq.Q(0, 1) | pq.Squeezing2(r=0.1, phi=np.pi / 3)
        pq.Q(1, 2) | pq.Squeezing2(r=0.2, phi=np.pi / 5)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program).state

    covariance_matrix = state.xpxp_covariance_matrix

    symplectic, diagonal = williamson(covariance_matrix, connector)

    assert is_diagonal(diagonal)
    assert is_symplectic(symplectic, form_func=xp_symplectic_form)
    assert np.all(np.isreal(symplectic))
    assert np.allclose(covariance_matrix, symplectic @ diagonal @ symplectic.T)


@pytest.mark.monkey
@pytest.mark.parametrize("connector", [NumpyConnector(), JaxConnector()])
def test_williamson_with_random_positive_definite_matrix(
    generate_random_positive_definite_matrix, connector
):
    dim = 4
    matrix = generate_random_positive_definite_matrix(dim)

    symplectic, diagonal = williamson(matrix, connector)

    assert is_diagonal(diagonal)
    assert is_symplectic(symplectic, form_func=xp_symplectic_form)
    assert np.all(np.isreal(symplectic))
    assert np.allclose(matrix, symplectic @ diagonal @ symplectic.T)


@pytest.mark.parametrize("connector", [NumpyConnector(), JaxConnector()])
def test_williamson_with_special_matrix(connector):
    """
    Sometimes scipy.linalg.sqrtm returns complex matrices even when real matrices with
    positive eigenvalues got specified. This test specifies a matrix for which the
    square root turns out to be complex even if it is real and positive definite.

    When a complex matrix is specified to the schur algorithm, automatically the complex
    schur algorithm is executed instead of the desired real schur algorithm.
    """
    matrix = np.array(
        [
            [
                1.8577097765113275,
                -0.03968353838339344,
                -0.13480023344314673,
                0.09038151421983714,
                -0.19128617316388555,
                0.061794725510946794,
                0.14126733590055737,
                -0.13966704763096188,
            ],
            [
                -0.03968353838339344,
                2.349487427421141,
                -0.07259190627416134,
                -0.01569343729182265,
                0.06179472551094684,
                0.05723660406987829,
                -0.07428264466311996,
                0.06923453621903403,
            ],
            [
                -0.13480023344314673,
                -0.07259190627416134,
                2.1222235731327235,
                0.21237578147380476,
                0.14126733590055734,
                -0.07428264466311996,
                0.09628621679115495,
                0.12278740167165433,
            ],
            [
                0.09038151421983714,
                -0.015693437291822643,
                0.21237578147380476,
                2.01240160261767,
                -0.13966704763096188,
                0.06923453621903401,
                0.12278740167165435,
                0.18926141391838203,
            ],
            [
                -0.19128617316388555,
                0.06179472551094682,
                0.14126733590055737,
                -0.13966704763096188,
                2.2225572459649756,
                0.03968353838339344,
                0.13480023344314673,
                -0.09038151421983714,
            ],
            [
                0.061794725510946814,
                0.05723660406987829,
                -0.07428264466311996,
                0.06923453621903403,
                0.03968353838339344,
                1.730779595055162,
                0.07259190627416134,
                0.015693437291822643,
            ],
            [
                0.14126733590055734,
                -0.07428264466311996,
                0.09628621679115495,
                0.12278740167165433,
                0.13480023344314673,
                0.07259190627416134,
                1.9580434493435799,
                -0.21237578147380476,
            ],
            [
                -0.13966704763096188,
                0.06923453621903401,
                0.12278740167165435,
                0.18926141391838203,
                -0.09038151421983714,
                0.015693437291822637,
                -0.21237578147380476,
                2.0678654198586335,
            ],
        ],
        dtype=float,
    )

    symplectic, diagonal = williamson(matrix, connector)

    assert is_diagonal(diagonal)
    assert is_symplectic(symplectic, form_func=xp_symplectic_form)
    assert np.all(np.isreal(symplectic))
    assert np.allclose(matrix, symplectic @ diagonal @ symplectic.T)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import DefaultDict, Optional, Tuple, Generator, Any, Dict, List

import abc
from collections import defaultdict
import numpy as np
from piquasso.api.config import Config
from piquasso.api.state import State
from piquasso.api.connector import BaseConnector

from piquasso._math.fock import get_fock_space_basis
from piquasso.api.exceptions import InvalidModes
from piquasso._simulators.plot import plot_wigner_function


class BaseFockState(State, abc.ABC):
    def __init__(
        self, *, d: int, connector: BaseConnector, config: Optional[Config] = None
    ) -> None:
        super().__init__(connector=connector, config=config)
        self._d = d
        cutoff = self._config.cutoff
        self._space = get_fock_space_basis(d=d, cutoff=cutoff)  # type: ignore

    @property
    def d(self) -> int:
        return self._d

    @property
    def norm(self) -> float:
        return self._connector.np.sum(self.fock_probabilities)

    def _as_code(self) -> str:
        return f"pq.Q() | pq.{self.__class__.__name__}(d={self.d})"

    @abc.abstractmethod
    def _get_empty(self) -> np.ndarray:
        """Returns a full-zero representation (i.e. norm 0)."""

    @abc.abstractmethod
    def _as_mixed(self):
        """Returns the density matrix of the Fock state."""

    @property
    @abc.abstractmethod
    def nonzero_elements(self) -> Generator[Tuple[complex, tuple], Any, None]:
        """The nonzero contributions to the state representation in Fock basis."""

    @property
    @abc.abstractmethod
    def density_matrix(self) -> np.ndarray:
        """The density matrix of the state in the truncated Fock space."""

    @abc.abstractmethod
    def reduced(self, modes: Tuple[int, ...]) -> "BaseFockState":
        """Reduces the state to a subsystem corresponding to the specified modes."""

    @abc.abstractmethod
    def normalize(self) -> None:
        """Normalizes the state to have norm 1."""

    @abc.abstractmethod
    def reset(self) -> None:
        """
        Resets the Fock state to a vacuum state.
        """

    @property
    def fock_amplitudes_map(self) -> DefaultDict[tuple, complex]:
        """Probability amplitudes of Fock basis states representing the state."""
        amplitude_map = defaultdict(complex)

        for elem in self.nonzero_elements:
            amplitude_map[elem[1]] = elem[0]

        return amplitude_map

    @property
    @abc.abstractmethod
    def fock_probabilities_map(self) -> Dict[Tuple[int, ...], float]:
        """The particle number detection probabilities in a map."""

    @abc.abstractmethod
    def quadratures_mean_variance(
        self, modes: Tuple[int, ...], phi: float = 0
    ) -> Tuple[float, float]:
        """Calculates the mean and the variance of the quadratures of a Fock State"""

    def wigner_function(
        self,
        positions: List[float],
        momentums: List[float],
        modes: Optional[Tuple[int, ...]] = None,
    ) -> np.ndarray:
        r"""
        This method calculates the Wigner function values at the specified position and
        momentum vectors, according to the following equation:

        .. math::
            W(r) = \frac{1}{\pi^d \sqrt{\mathrm{det} \sigma}}
                \exp \big (
                    - (r - \mu)^T
                    \sigma^{-1}
                    (r - \mu)
                \big ).

        Note:
            The implementation is copied from
            `QuTiP <https://qutip.org/docs/latest/modules/qutip/wigner.html#wigner>`_.

        Note:
            Only single modes are supported.

        Args:
            positions (list[float]): List of position vectors.
            momentums (list[float]): List of momentum vectors.
            modes (tuple[int], optional):
                Modes where Wigner function should be calculcated.

        Returns:
            numpy.ndarray:
                The Wigner function values in the shape of a grid specified by the
                input.
        """

        # Since this code is copied from QuTiP, this copyright notice is inserted here:

        #    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
        #    All rights reserved.
        #
        #    Redistribution and use in source and binary forms, with or without
        #    modification, are permitted provided that the following conditions are
        #    met:
        #
        #    1. Redistributions of source code must retain the above copyright notice,
        #       this list of conditions and the following disclaimer.
        #
        #    2. Redistributions in binary form must reproduce the above copyright
        #       notice, this list of conditions and the following disclaimer in the
        #       documentation and/or other materials provided with the distribution.
        #
        #    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
        #       of its contributors may be used to endorse or promote products derived
        #       from this software without specific prior written permission.
        #
        #    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
        #    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
        #    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
        #    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
        #    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
        #    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
        #    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
        #    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
        #    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
        #    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
        #    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

        if self._config.validate and self.d != 1 and (modes is None or len(modes) != 1):
            raise InvalidModes(
                "The Wigner function can only be calculated for a single mode: "
                f"modes={modes}."
            )

        state = self if modes is None else self.reduced(modes)

        rho = state.density_matrix

        g = np.sqrt(2 / self._config.hbar)

        # QuTiP implementation starts from here

        M = np.prod(rho.shape[0])
        X, Y = np.meshgrid(positions, momentums)
        A = 0.5 * g * (X + 1.0j * Y)

        Wlist = np.array(
            [np.zeros(np.shape(A), dtype=self._config.complex_dtype) for k in range(M)]
        )
        Wlist[0] = np.exp(-2.0 * abs(A) ** 2) / np.pi

        W = np.real(rho[0, 0]) * np.real(Wlist[0])
        for n in range(1, M):
            Wlist[n] = (2.0 * A * Wlist[n - 1]) / np.sqrt(n)
            W += 2 * np.real(rho[0, n] * Wlist[n])

        for m in range(1, M):
            temp = np.copy(Wlist[m])
            Wlist[m] = (2 * np.conj(A) * temp - np.sqrt(m) * Wlist[m - 1]) / np.sqrt(m)

            # Wlist[m] = Wigner function for |m><m|
            W += np.real(rho[m, m] * Wlist[m])

            for n in range(m + 1, M):
                temp2 = (2 * A * Wlist[n - 1] - np.sqrt(m) * temp) / np.sqrt(n)
                temp = np.copy(Wlist[n])
                Wlist[n] = temp2

                # Wlist[n] = Wigner function for |m><n|
                W += 2 * np.real(rho[m, n] * Wlist[n])

        return 0.5 * W * g**2

    def plot_wigner(
        self,
        positions: List[float],
        momentums: List[float],
        mode: Optional[int] = None,
    ) -> None:
        r"""
        Plots the Wigner function in phase space for a single mode using the
        given positions and momentums.

        Args:
            positions (List[float]): List of position values (x-axis)
            momentums (List[float]): List of momentum values (p-axis)
            mode (int, optional):
                Mode where Wigner function should be calculcated.

        Note:
            Only a single mode is supported.

        Returns:
            None: This method generates a plot and does not return a value.

        """
        if self._config.validate and mode is not None and mode > self.d - 1:
            raise InvalidModes(
                f"Mode {mode} is out of range for the state with {self.d} modes."
            )

        if mode is not None:
            mode_tuple = (mode,)
        else:
            mode_tuple = None

        if self._config.validate and self.d != 1 and mode_tuple is None:
            raise InvalidModes(
                "The Wigner function can only be plotted for a single mode: "
                f"modes={mode_tuple} was specified."
            )

        fock_wigner_function_values = self.wigner_function(
            positions=positions, momentums=momentums, modes=mode_tuple
        )
        x, p = np.meshgrid(positions, momentums)
        xlist = x.tolist()
        plist = p.tolist()

        plot_wigner_function(
            fock_wigner_function_values,
            positions=xlist,
            momentums=plist,
        )

    def fidelity(self, state: "BaseFockState") -> float:
        r"""Calculates the state fidelity between two quantum states.

        The state fidelity :math:`F` between two density matrices
        :math:`\rho_1, \rho_2` is given by:

        .. math::
            \operatorname{F}(\rho_1, \rho_2) = \operatorname{Tr}(\sqrt{\sqrt{\rho_1}
                \rho_2\sqrt{\rho_1}})^2 = \operatorname{Tr}(\sqrt{\rho_1 \rho_2})^2

        Args:
            state:
                Either a :class:`~piquasso._simulators.fock.pure.state.PureFockState`
                or a :class:`~piquasso._simulators.fock.general.state.FockState` that
                can be used to calculate the fidelity aganist it.

        Returns:
            float: The calculated fidelity.
        """
        geometric_mean = np.sqrt(
            np.linalg.eigvals(self.density_matrix @ state.density_matrix)
        )
        return np.abs(np.sum(geometric_mean)) ** 2

    def get_marginal_fock_probabilities(self, modes: Tuple[int, ...]) -> np.ndarray:
        """Return the particle number probabilities on a given subsystem.

        Args:
            modes (tuple[int, ...]): The indices of the modes to keep.

        Returns:
            numpy.ndarray: The marginal particle number detection probabilities.
        """

        return self.reduced(modes).fock_probabilities

    @abc.abstractmethod
    def get_purity(self):
        """The purity of the Fock state."""



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq

from piquasso.api.exceptions import PiquassoException


def test_registering_instruction_by_subclassing():
    class OtherBeamsplitter(pq.Instruction):
        pass

    assert pq.Instruction.get_subclass("OtherBeamsplitter") is OtherBeamsplitter


def test_subclassing_instruction_with_existing_name_is_successful():
    class Beamsplitter(pq.Instruction):
        pass

    assert pq.Instruction.get_subclass("Beamsplitter") is Beamsplitter
    assert pq.Beamsplitter is not Beamsplitter

    # Teardown
    pq.Instruction.set_subclass(pq.Beamsplitter)
    assert pq.Instruction.get_subclass("Beamsplitter") is pq.Beamsplitter


def test_set_subclass_with_no_subclass():
    any_other_class = object

    with pytest.raises(PiquassoException):
        pq.Instruction.set_subclass(any_other_class)
